__author__ = 'Robbert Harms'
__date__ = "2014-06-19"
__license__ = "LGPL v3"
__maintainer__ = "Robbert Harms"
__email__ = "robbert.harms@maastrichtuniversity.nl"


class AbstractParameterProposal(object):
    """The parameter proposals are meant for use during sampling.

    They indicate how a proposal should be generated for the given parameter.

    The proposals can allow for dynamic updates to better adapt to the landscape during sampling.
    """

    @property
    def is_symmetric(self):
        """Check if this proposal is symmetric. That is, if q(x|y) == q(y|x)."""
        return True

    def get_parameters(self):
        """The proposal parameters.

        Returns:
            list of ProposalParameter: the proposal parameter objects used by this proposal
        """

    def get_proposal_function(self):
        """Get the proposal function as a CL string. This should include include guards (#ifdef's).

        This should follow the signature:
        mot_float_type <proposal_fname>(mot_float_type current, void* rng_data, <additional_parameters>)

        That is, it can have more than two parameter, but the first two are obligatory. The additional parameters
        are defined by the get_parameters function of this python class.

        Returns:
            str: The cl function
        """

    def get_proposal_function_name(self):
        """Get the name of the proposal function call.

         This is used by the model builder to construct the call to the proposal function.

         Returns:
            str: name of the function
        """

    def get_proposal_logpdf_function(self):
        """Get the proposal pdf function as a CL string.

        This function is used if the proposal is not symmetric. The implementation should include include guards.

        This should follow the signature:
            mot_float_type <proposal_pdf_fname>(mot_float_type proposal, mot_float_type current,
                                                <additional_parameters>)

        Returns:
            str: The proposal log pdf function as a CL string
        """

    def get_proposal_logpdf_function_name(self):
        """Get the name of the proposal logpdf function call.

         This is used by the model builder to construct the call to the proposal logpdf function.

         Returns:
            str: name of the function
        """


class ProposalParameter(object):

    def __init__(self, default_value, adaptable):
        """Container class for parameters of a proposal function.

        Args:
            default_value (double): the parameter value
            adaptable (boolean): if this parameter is adaptable during sampling

        Attributes:
            default_value (double): the parameter value
            adaptable (boolean): if this parameter is adaptable
        """
        self.default_value = default_value
        self.adaptable = adaptable

    def get_parameter_update_function(self):
        """Get the parameter update function used to update this proposal parameter.

        Returns:
            A function with the signature:
                mot_float_type <func_name>(const mot_float_type current_value,
                                           const uint acceptance_counter,
                                           const uint jump_counter)

            Where current value is the current value for this proposal parameter, acceptance counter is the
            number of accepted steps and jump counter the number of jumps.
        """
        return '''
            #ifndef PROPOSAL_DEFAULT_PARAMETER_UPDATE
            #define PROPOSAL_DEFAULT_PARAMETER_UPDATE

            mot_float_type proposal_default_parameter_update(const mot_float_type current_value,
                                                             const uint acceptance_counter,
                                                             const uint jump_counter){
                return min(current_value *
                            sqrt( (mot_float_type)(acceptance_counter+1) /
                                  ((jump_counter - acceptance_counter) + 1)
                            ),
                           (mot_float_type)1e10);
            }

            #endif //PROPOSAL_DEFAULT_PARAMETER_UPDATE
        '''

    def get_parameter_update_function_name(self):
        return 'proposal_default_parameter_update'


class GaussianProposal(AbstractParameterProposal):

    def __init__(self, std=1.0, adaptable=True):
        """Create a new proposal function using a Gaussian distribution with the given scale.

        Args:
            std (float): The scale of the Gaussian distribution.
            adaptable (boolean): If this proposal is adaptable during sampling
        """
        self._parameters = [ProposalParameter(std, adaptable)]

    def get_parameters(self):
        return self._parameters

    def get_proposal_function(self):
        return '''
            #ifndef PROP_GAUSSIANPROPOSAL_CL
            #define PROP_GAUSSIANPROPOSAL_CL

            mot_float_type proposal_gaussianProposal(mot_float_type current,
                                                     void* rng_data,
                                                     mot_float_type std){
                return fma(std, (mot_float_type)frandn(rng_data), current);
            }

            #endif //PROP_GAUSSIANPROPOSAL_CL
        '''

    def get_proposal_function_name(self):
        return 'proposal_gaussianProposal'

    def get_proposal_logpdf_function(self):
        return '''
            #ifndef PROP_GAUSSIANPROPOSALLOGPDF_CL
            #define PROP_GAUSSIANPROPOSALLOGPDF_CL

            mot_float_type proposal_gaussianProposalLogPDF(mot_float_type x,
                                                           mot_float_type mu,
                                                           mot_float_type std){
                return log(std * sqrt(2 * M_PI)) - (((x - mu) * (x - mu)) / (2 * std * std));
            }

            #endif //PROP_GAUSSIANPROPOSALLOGPDF_CL
        '''

    def get_proposal_logpdf_function_name(self):
        return 'proposal_gaussianProposalLogPDF'


class CircularGaussianProposal(AbstractParameterProposal):

    def __init__(self, modulus, std=1.0, adaptable=True):
        """A Gaussian distribution which loops around the given modulus.

        Args:
            modulus (float): at which point we loop around
            std (float): The scale of the Gaussian distribution.
            adaptable (boolean): If this proposal is adaptable during sampling
        """
        self._parameters = [ProposalParameter(std, adaptable)]
        self.modulus = modulus

    def get_parameters(self):
        return self._parameters

    def get_proposal_function(self):
        return '''
            #ifndef PROP_CIRCULARGAUSSIANPROPOSAL_CL
            #define PROP_CIRCULARGAUSSIANPROPOSAL_CL

            mot_float_type proposal_circular_gaussianProposal(mot_float_type current,
                                                              void* rng_data,
                                                              mot_float_type std){
                double x1 = fma(std, (mot_float_type)frandn(rng_data), current);
                double x2 = ''' + str(self.modulus) + ''';
                return (mot_float_type) (x1 - floor(x1 / x2) * x2);
            }

            #endif //PROP_CIRCULARGAUSSIANPROPOSAL_CL
        '''

    def get_proposal_function_name(self):
        return 'proposal_circular_gaussianProposal'

    def get_proposal_logpdf_function(self):
        return '''
            #ifndef PROP_CIRCULARGAUSSIANPROPOSALLOGPDF_CL
            #define PROP_CIRCULARGAUSSIANPROPOSALLOGPDF_CL

            mot_float_type proposal_circular_gaussianProposalLogPDF(mot_float_type x,
                                                                    mot_float_type mu,
                                                                    mot_float_type std){
                return log(std * sqrt(2 * M_PI)) - (((x - mu) * (x - mu)) / (2 * std * std));
            }

            #endif //PROP_CIRCULARGAUSSIANPROPOSALLOGPDF_CL
        '''

    def get_proposal_logpdf_function_name(self):
        return 'proposal_circular_gaussianProposalLogPDF'




import logging
import numbers
import numpy as np
import pyopencl as cl
from mot.cl_routines.base import CLRoutine
from mot.load_balance_strategies import Worker


__author__ = 'Robbert Harms'
__date__ = "2014-04-26"
__license__ = "LGPL v3"
__maintainer__ = "Robbert Harms"
__email__ = "robbert.harms@maastrichtuniversity.nl"


class AbstractFilter(CLRoutine):

    def __init__(self, size, cl_environments=None, load_balancer=None):
        """Initialize the filter routine.

        Args:
            size (int or tuple): Either a single dimension size for all dimensions or one value
                for each dimension of the input data to the filter function. Maximum number of dimensions is 3.
                Either way this value is the distance to the left and to the right of each value.
                That means that the total kernel size is the product of 1 + 2*s for each size s of each dimension.
            cl_environments: The cl environments
            load_balancer: The load balancer to use

        Attributes:
            size (int or tuple): (x, y, z, ...). Either a single dimension size for all dimensions or one value
                for each dimension of the input data to the filter function.
                Either way this value is the distance to the left and to the right of each value.
                That means that the total kernel size is the product of 1 + 2*s for each size s of each dimension.
        """
        super(AbstractFilter, self).__init__(cl_environments=cl_environments, load_balancer=load_balancer)
        self.size = size
        self._logger = logging.getLogger(__name__)

    def filter(self, value, mask=None, double_precision=False, nmr_of_times=1):
        """Filter the given volumes in the given dictionary.

        If a dict is given as a value the filtering is applied to every value in the dictionary. This can be spread
        over the different devices. If a single ndarray is given the filtering is performed only on one device.

        Args:
            value (dict or array like): an single array to filter (dimensions must match the size specified in
                the constructor). Can also be a dictionary with a list of ndarrays.
            mask (array like): A single array of the same dimension as the input value. This can be used to
                mask values from being used by the filtering routines. They are not used for filtering other values
                and are not filtered themselves.
            double_precision (boolean): if we will use double or float
            nmr_of_times (int): how many times we would like to repeat the filtering (per input volume).

        Returns:
            The same type as the input. A new set of volumes with the same keys, or a single array. All filtered.
        """
        if nmr_of_times < 1:
            nmr_of_times = 1

        if nmr_of_times == 1:
            self._logger.info('Applying filtering with filter {0}'.format(self.__class__.__name__))
            if isinstance(value, dict):
                return self._filter(value, mask, double_precision)
            else:
                return self._filter({'value': value}, mask, double_precision)['value']
        else:
            filtered = self.filter(value, mask, double_precision, 1)
            return self.filter(filtered, mask, double_precision, nmr_of_times - 1)

    def _filter(self, volumes_dict, mask, double_precision):
        results_dict = {}

        np_dtype = np.float32
        if double_precision:
            np_dtype = np.float64

        for key, value in volumes_dict.items():
            if len(value.shape) > 3 and value.shape[3] > 1:
                raise ValueError('The given volume {} is a 4d volume with a 4th dimension >1. We can not use this.')

            volumes_dict[key] = value.astype(dtype=np_dtype, copy=False, order='C')
            results_dict[key] = np.zeros_like(volumes_dict[key], dtype=np_dtype, order='C')

        if mask is not None:
            mask = mask.astype(np.int8, order='C', copy=True)

        volumes_list = list(volumes_dict.items())
        workers = self._create_workers(self._get_worker_generator(self, results_dict, volumes_list, mask,
                                                                  double_precision))
        self._load_balancer.process(workers, len(volumes_list))

        return results_dict

    def _get_worker_generator(self, *args):
        """Generate the worker generator callback for the function _create_workers()

        This is supposed to be overwritten by the implementing filterer.

        Returns:
            the python callback for generating the worker
        """
        return lambda cl_environment: AbstractFilterWorker(cl_environment, self.get_compile_flags_list(), *args)


class AbstractFilterWorker(Worker):

    def __init__(self, cl_environment, compile_flags, parent_filter, results_dict,
                 volumes_list, mask, double_precision):
        """Create a filter worker.

        Args:
            nmr_of_times: the number of times we want to apply the filter per dataset.
        """

        super(AbstractFilterWorker, self).__init__(cl_environment)
        self._parent_filter = parent_filter
        self._size = self._parent_filter.size
        self._results_dict = results_dict
        self._volumes_list = volumes_list
        self._volume_shape = volumes_list[0][1].shape[0:3]
        self._double_precision = double_precision
        self._mask = mask
        self._logger = logging.getLogger(__name__)

        if mask is None:
            self._use_mask = False
        else:
            self._use_mask = True
            self._mask_buf = cl.Buffer(self._cl_run_context.context,
                                       cl.mem_flags.READ_ONLY | cl.mem_flags.COPY_HOST_PTR,
                                       hostbuf=self._mask)

        self._kernel = self._build_kernel(compile_flags)

    def calculate(self, range_start, range_end):
        volumes_to_run = [self._volumes_list[i] for i in range(len(self._volumes_list)) if range_start <= i < range_end]

        volume_buf = cl.Buffer(self._cl_run_context.context,
                               cl.mem_flags.READ_ONLY | cl.mem_flags.COPY_HOST_PTR,
                               hostbuf=volumes_to_run[0][1])

        results_buf = cl.Buffer(self._cl_run_context.context,
                                cl.mem_flags.WRITE_ONLY | cl.mem_flags.COPY_HOST_PTR,
                                hostbuf=self._results_dict[volumes_to_run[0][0]])

        event = None
        for key, value in volumes_to_run:
            cl.enqueue_copy(self._cl_run_context.queue, volume_buf, value, is_blocking=False)
            cl.enqueue_copy(self._cl_run_context.queue, results_buf, self._results_dict[key], is_blocking=False)

            buffers = [volume_buf]
            if self._use_mask:
                buffers.append(self._mask_buf)
            buffers.append(results_buf)

            event = self._kernel.filter(self._cl_run_context.queue, self._volume_shape, None, *buffers)
            event = cl.enqueue_copy(self._cl_run_context.queue, self._results_dict[key], results_buf,
                                    is_blocking=False, wait_for=[event])

        return [event]

    def _get_kernel_source(self):
        """Get the kernel source for this filtering kernel.

           This should be implemented by the subclass.
        """

    def _get_size_in_dimension(self, dimension):
        if isinstance(self._size, numbers.Number):
            return self._size
        else:
            return self._size[dimension]

    def _calculate_kernel_size_in_dimension(self, dimension):
        return self._get_size_in_dimension(dimension) * 2 + 1

    def _get_ks_dimension_inits(self, nmr_dimensions):
        """Get the kernel source part for the dimension initializations"""
        s = ''
        for i in range(nmr_dimensions):
            s += 'int dim' + str(i) + ' = get_global_id(' + str(i) + ');' + "\n"
        return s

    def _get_ks_sub2ind_func(self, volume_shape):
        """Get the kernel source part for converting array subscripts to indices"""
        s = 'int sub2ind('
        for i in range(len(volume_shape)):
            s += 'const int dim' + str(i) + ', '
        s = s[0:-2] + '){' + "\n"
        s += 'return '
        for i, d in enumerate(volume_shape):
            stride = ''
            for ds in volume_shape[(i + 1):]:
                stride += ' * ' + str(ds)
            s += 'dim' + str(i) + stride + ' + '
        s = s[0:-3] + ';' + "\n"
        s += '}' + "\n"
        return s

    def _get_ks_sub2ind_func_call(self, nmr_dimensions):
        """Get the kernel source part for the function call for converting array subscripts to indices"""
        s = 'sub2ind('
        for i in range(nmr_dimensions):
            s += 'dim' + str(i) + ', '
        return s[0:-2] + ')'

    def _calculate_length(self, nmr_dimensions):
        """Calculate the length of the array given the number of dimensions.

        The kernel size is determined by the global size tuple. For each dimension this specifies the number of
        values we look to the right and to the left to calculate the new value. This means that the total kernel size
        of all dimensions together is the multiplication of 2 * n + 1 for each dimension: left + right + current.

        Args:
            nmr_dimensions (int): the number of dimensions
        """
        n = 1
        for dimension in range(nmr_dimensions):
            s = self._get_size_in_dimension(dimension)
            n *= (s * 2 + 1)
        return n

    def _get_ks_dimension_sizes(self, volume_shape):
        """Get the kernel source for the start and end of each of the dimensions"""
        s = ''
        for i, d in enumerate(volume_shape):
            s += 'int dim' + str(i) + '_start = max(0, dim' + str(i) + ' - ' + str(self._get_size_in_dimension(i)) \
                 + ');' + "\n"
            s += 'int dim' + str(i) + '_end = min(' + str(d) + ', dim' + str(i) + ' + ' + \
                 str(self._get_size_in_dimension(i)) + ' + 1);' + "\n"
        return s



from mot.model_building.cl_functions.parameters import FreeParameter
from mot.model_building.cl_functions.base import ModelFunction
from mot.model_building.cl_functions.library_functions import Bessel
from mot.model_building.parameter_functions.transformations import ClampTransform
from mot.cl_data_type import CLDataType


__author__ = 'Robbert Harms'
__date__ = "2014-08-05"
__license__ = "LGPL v3"
__maintainer__ = "Robbert Harms"
__email__ = "robbert.harms@maastrichtuniversity.nl"


class EvaluationModel(ModelFunction):

    def __init__(self, name, cl_function_name, parameter_list, dependency_list=()):
        """The evaluation model is the model under which you evaluate the estimated results against the data.

        This normally embed the noise model assumptions of your data.
        """
        super(EvaluationModel, self).__init__(name, cl_function_name, parameter_list, dependency_list=dependency_list)

    def get_objective_function(self, fname, inst_per_problem, eval_fname, obs_fname, param_listing):
        """Get the cl code for the objective function under the given noise model.

        Args:
            fname (str): the name of the resulting function
            inst_per_problem (int): the number of instances per problem
            eval_fname (str): the name of the function that can be called to get the evaluation, its signature is:

                .. code-block:: c

                    mot_float_type <fname>(const optimize_data* data, const mot_float_type* x,
                                           const int observation_index);

            obs_fname (str): the name of the function that can be called for the observed data, its signature is:

                .. code-block:: c

                    mot_float_type <fname>(const optimize_data* data, const int observation_index);

            param_listing (str): the parameter listings for the parameters of the noise model

        Returns:
            str: The objective function under this noise model, its signature is:

                .. code-block:: c

                    double <fname>(const optimize_data* const data, mot_float_type* const x);

            That is, it always returns a double since the summations may get large.
        """

    def get_objective_list_function(self, fname, inst_per_problem, eval_fname, obs_fname, param_listing):
        """Get the cl code for the objective function list under the given noise model.

        This function is used by some evaluation routines (like for example LevenbergMarquardt) that need
        a list of objective values (one per instance point), instead of a single objective function scalar.

        The returned CL function should accept a result vector in which the results per instance point ought
        to be placed.

        Args:
            fname (str): the name of the resulting function
            inst_per_problem (int): the number of instances per problem
            eval_fname (str): the name of the function that can be called to get the evaluation, its signature is:

                .. code-block:: c

                    mot_float_type <fname>(const optimize_data* data, const mot_float_type* x,
                                           const int observation_index);

            obs_fname (str): the name of the function that can be called for the observed data, its signature is:

                .. code-block:: c

                    mot_float_type <fname>(const optimize_data* data, const int observation_index);

            param_listing (str): the parameter listings for the parameters of the noise model

        Returns:
            str: The objective function under this noise model, its signature is:

                .. code-block:: c

                    void (const optimize_data* const data, mot_float_type* const x, mot_float_type* result);
        """

    def get_log_likelihood_function(self, fname, inst_per_problem, eval_fname, obs_fname, param_listing,
                                    full_likelihood=True):
        """Get the cl code for the log likelihood function under the given noise model.

        Args:
            fname (str): the name of the resulting function
            inst_per_problem (int): the number of instances per problem
            eval_fname (str): the name of the function that can be called to get the evaluation, its signature is:

                .. code-block:: c

                    mot_float_type <fname>(const optimize_data* data, const mot_float_type* x,
                                           const int observation_index);

            obs_fname (str): the name of the function that can be called for the observed data, its signature is:

                .. code-block:: c

                    mot_float_type <fname>(const optimize_data* data, const int observation_index);

            param_listing (str): the parameter listings for the parameters of the noise model
            full_likelihood (boolean): if we want the complete likelihood, or if we can drop the constant terms.
                The default is the complete likelihood. Disable for speed.

        Returns:
            str: the objective function under this noise model, its signature is:

                .. code-block:: c

                    double <fname>(const optimize_data* const data, mot_float_type* const x);

            That is, it always returns a double since the summations may get large.
        """

    def set_noise_level_std(self, noise_std):
        """Set the estimate of the noise level standard deviation.

        We put this here as a method to make the method work with object oriented polymorphism. That is, not
        all noiselevels use the same parameter for the noise standard deviation.

        This method makes no assumptions about the state of the parameters affected, they can be fixed or not. This
        function should just set the right value.

        Args:
            noise_std (double): the noise standard deviation

        Returns:
            self: for chaining
        """
        return self

    def get_noise_level_std(self):
        """Get the estimate of the noise level standard deviation.

        Returns:
            noise_std (double): the noise standard deviation
        """
        return 1


class SumOfSquares(EvaluationModel):

    def __init__(self):
        """Evaluates the distance between the estimated signal and the data using the sum of squared distance.

        This is implemented as::

            sum((observation - evaluation)^2)
        """
        super(EvaluationModel, self).__init__('SumOfSquaresNoise', 'sumOfSquaresNoise', (), ())

    def get_objective_function(self, fname, inst_per_problem, eval_fname, obs_fname, param_listing):
        return '''
            double ''' + fname + '''(const optimize_data* const data, mot_float_type* const x){
                ''' + param_listing + '''
                double sum = 0.0;
                for(int i = 0; i < ''' + str(inst_per_problem) + '''; i++){
                    sum += pown(''' + obs_fname + '''(data, i) - ''' + eval_fname + '''(data, x, i), 2);
                }
                return sum;
            }
        '''

    def get_log_likelihood_function(self, fname, inst_per_problem, eval_fname, obs_fname, param_listing,
                                    full_likelihood=True):
        return '''
            double ''' + fname + '''(const optimize_data* const data, const mot_float_type* const x){
                ''' + param_listing + '''
                double sum = 0.0;
                for(int i = 0; i < ''' + str(inst_per_problem) + '''; i++){
                    sum += pown(''' + obs_fname + '''(data, i) - ''' + eval_fname + '''(data, x, i), 2);
                }
                return - sum;
            }
        '''

    def get_objective_list_function(self, fname, inst_per_problem, eval_fname, obs_fname, param_listing):
        return '''
            void ''' + fname + '''(const optimize_data* const data, mot_float_type* const x, mot_float_type* result){
                ''' + param_listing + '''
                for(int i = 0; i < ''' + str(inst_per_problem) + '''; i++){
                    result[i] = ''' + obs_fname + '''(data, i) - ''' + eval_fname + '''(data, x, i);
                }
            }
        '''


class GaussianEvaluationModel(EvaluationModel):

    def __init__(self):
        """This uses the log of the Gaussian PDF for the maximum likelihood estimator and for the log likelihood.

        The PDF is defined as:

        .. code-block:: c

            PDF = 1/(sigma * sqrt(2*pi)) * exp(-(observation - evaluation)^2 / (2 * sigma^2))

        To have the joined probability over all instances one would normally have to take the product
        over all ``n`` instances:

        .. code-block:: c

            product(PDF)

        Instead of taking the product of this PDF we take the sum of the log of the PDF:

        .. code-block:: c

            sum(log(PDF))

        Where the log of the PDF is given by:

        .. code-block:: c

            log(PDF) = - ((observation - evaluation)^2 / (2 * sigma^2)) - log(sigma * sqrt(2*pi))


        For the maximum likelihood estimator we then need to use the negative of this sum:

        .. code-block:: c

            - sum(log(PDF)).
        """
        super(GaussianEvaluationModel, self).__init__(
            'GaussianNoise',
            'gaussianNoiseModel',
            (FreeParameter(CLDataType.from_string('mot_float_type'), 'sigma', True, 1, 0, 'INFINITY',
                           parameter_transform=ClampTransform()),), ())

    def get_objective_function(self, fname, inst_per_problem, eval_fname, obs_fname, param_listing):
        """Get the Gaussian objective function.

        This omits the constant terms for speed reasons. Omitted terms are:

         .. code-block:: c

            + log(GaussianNoise_sigma * sqrt(2 * M_PI))

        """
        return '''
            double ''' + fname + '''(const optimize_data* const data, mot_float_type* const x){
                ''' + param_listing + '''
                double sum = 0.0;
                for(int i = 0; i < ''' + str(inst_per_problem) + '''; i++){
                    sum += pown(''' + obs_fname + '''(data, i) - ''' + eval_fname + '''(data, x, i), 2);
                }
                return sum / (2 * GaussianNoise_sigma * GaussianNoise_sigma);
            }
        '''

    def get_objective_list_function(self, fname, inst_per_problem, eval_fname, obs_fname, param_listing):
        return '''
            void ''' + fname + '''(const optimize_data* const data, mot_float_type* const x, mot_float_type* result){
                ''' + param_listing + '''
                for(int i = 0; i < ''' + str(inst_per_problem) + '''; i++){
                    result[i] = ''' + obs_fname + '''(data, i) - ''' + eval_fname + '''(data, x, i);
                }
            }
        '''

    def get_log_likelihood_function(self, fname, inst_per_problem, eval_fname, obs_fname, param_listing,
                                    full_likelihood=True):
        return '''
            double ''' + fname + '''(const optimize_data* const data, const mot_float_type* const x){
                ''' + param_listing + '''
                double sum = 0.0;
                for(int i = 0; i < ''' + str(inst_per_problem) + '''; i++){
                    sum += pown(''' + obs_fname + '''(data, i) - ''' + eval_fname + '''(data, x, i), 2);
                }
                return - sum / (2 * GaussianNoise_sigma * GaussianNoise_sigma)
                    ''' + ('+' + str(inst_per_problem) + ' * log(GaussianNoise_sigma * sqrt(2 * M_PI))'
                           if full_likelihood else '') + ''';

            }
        '''

    def set_noise_level_std(self, noise_std):
        self.parameter_list[0].value = noise_std
        return self

    def get_noise_level_std(self):
        return self.parameter_list[0].value


class OffsetGaussianEvaluationModel(EvaluationModel):

    def __init__(self):
        """This uses the log of the Gaussian PDF for the maximum likelihood estimator and for the log likelihood.

        The PDF is defined as:

        .. code-block:: c

            PDF = 1/(sigma * sqrt(2*pi)) * exp(-(observation - sqrt(evaluation^2 + sigma^2))^2 / (2 * sigma^2))

        To have the joined probability over all instances one would have to take the product over all n instances:

        .. code-block:: c

            product(PDF)

        Instead of taking the product of this PDF we take the sum of the log of the PDF:

        .. code-block:: c

            sum(log(PDF))

        Where the log of the PDF is given by:

        .. code-block:: c

            log(PDF) = - ((observation - sqrt(evaluation^2 + sigma^2))^2 / (2 * sigma^2)) - log(sigma * sqrt(2*pi))

        For the maximum likelihood estimator we use the negative of this sum:

        .. code-block:: c

            -sum_n(log(PDF)).
        """
        super(OffsetGaussianEvaluationModel, self).__init__(
            'OffsetGaussianNoise',
            'offsetGaussianNoiseModel',
            (FreeParameter(CLDataType.from_string('mot_float_type'), 'sigma', True, 1, 0, 'INFINITY',
                           parameter_transform=ClampTransform()),), ())

    def get_objective_function(self, fname, inst_per_problem, eval_fname, obs_fname, param_listing):
        """Get the Offset Gaussian objective function.

        This omits the constant terms for speed reasons. Omitted terms are:

         .. code-block:: c

            (+ log(OffsetGaussianNoise_sigma * sqrt(2 * M_PI)))
        """
        return '''
            double ''' + fname + '''(const optimize_data* const data, mot_float_type* const x){
                ''' + param_listing + '''
                double sum = 0.0;
                for(int i = 0; i < ''' + str(inst_per_problem) + '''; i++){
                    sum += pown(''' + obs_fname + '''(data, i) -
                                sqrt(pown(''' + eval_fname + '''(data, x, i), 2) +
                                    (OffsetGaussianNoise_sigma * OffsetGaussianNoise_sigma)), 2);
                }
                return sum / (2 * pown(OffsetGaussianNoise_sigma, 2));
            }
        '''

    def get_objective_list_function(self, fname, inst_per_problem, eval_fname, obs_fname, param_listing):
        return '''
            void ''' + fname + '''(const optimize_data* const data, mot_float_type* const x, mot_float_type* result){
                ''' + param_listing + '''
                for(int i = 0; i < ''' + str(inst_per_problem) + '''; i++){
                    result[i] = ''' + obs_fname + '''(data, i) -
                                    sqrt(pown(''' + eval_fname + '''(data, x, i), 2)
                                            + (OffsetGaussianNoise_sigma * OffsetGaussianNoise_sigma));
                }
            }
        '''

    def get_log_likelihood_function(self, fname, inst_per_problem, eval_fname, obs_fname, param_listing,
                                    full_likelihood=True):
        return '''
            double ''' + fname + '''(const optimize_data* const data, mot_float_type* const x){
                ''' + param_listing + '''
                double sum = 0.0;
                for(int i = 0; i < ''' + str(inst_per_problem) + '''; i++){
                    sum += pown(''' + obs_fname + '''(data, i) -
                                sqrt(pown(''' + eval_fname + '''(data, x, i), 2) +
                                    (OffsetGaussianNoise_sigma * OffsetGaussianNoise_sigma)), 2);
                }
                return - sum / (2 * pown(OffsetGaussianNoise_sigma, 2))
                    ''' + ('+' + str(inst_per_problem) + ' * log(OffsetGaussianNoise_sigma * sqrt(2 * M_PI))'
                           if full_likelihood else '') + ''';
            }
        '''

    def set_noise_level_std(self, noise_std):
        self.parameter_list[0].value = noise_std
        return self

    def get_noise_level_std(self):
        return self.parameter_list[0].value


class RicianEvaluationModel(EvaluationModel):

    def __init__(self):
        """This uses the log of the Rice PDF for the maximum likelihood estimator and for the log likelihood.

        The PDF is defined as:

        .. code-block:: c

            PDF = (observation/sigma^2)
                    * exp(-(observation^2 + evaluation^2) / (2 * sigma^2))
                    * bessel_i0((observation * evaluation) / sigma^2)

        Where where ``bessel_i0(z)`` is the modified Bessel function of the first kind with order zero. To have the
        joined probability over all instances one would have to take the product over all n instances:

        .. code-block:: c

            product(PDF)

        Instead of taking the product of this PDF over all instances we take the sum of the log of the PDF:

        .. code-block:: c

            sum(log(PDF))

        Where the log of the PDF is given by:

        .. code-block:: c

            log(PDF) = log(observation/sigma^2)
                        - (observation^2 + evaluation^2) / (2 * sigma^2)
                        + log(bessel_i0((observation * evaluation) / sigma^2))

        For the maximum likelihood estimator we use the negative of this sum:

        .. code-block:: c

            -sum(log(PDF)).
        """
        super(RicianEvaluationModel, self).__init__(
            'RicianNoise',
            'ricianNoiseModel',
            (FreeParameter(CLDataType.from_string('mot_float_type'), 'sigma', True, 1, 0, 'INFINITY',
                           parameter_transform=ClampTransform()),),
            (Bessel(),))

    def get_objective_function(self, fname, inst_per_problem, eval_fname, obs_fname, param_listing):
        """Get the Rician objective function.

        This omits the constant terms for speed reasons. Omitted terms are:

         .. code-block:: c

            + log(observation / (RicianNoise_sigma * RicianNoise_sigma))
            - ((observation * observation) / (2 * (RicianNoise_sigma * RicianNoise_sigma)))

        """
        return '''
            double ''' + fname + '''(const optimize_data* const data, mot_float_type* const x){
                ''' + param_listing + '''
                double sum = 0.0;
                double observation;
                double evaluation;
                for(int i = 0; i < ''' + str(inst_per_problem) + '''; i++){
                    observation = (double)''' + obs_fname + '''(data, i);
                    evaluation = (double)''' + eval_fname + '''(data, x, i);

                    sum +=  - ((evaluation * evaluation) / (2 * RicianNoise_sigma * RicianNoise_sigma))
                            + log_bessel_i0((observation * evaluation) / (RicianNoise_sigma * RicianNoise_sigma));
                }
                return -sum;
            }
        '''

    def get_objective_list_function(self, fname, inst_per_problem, eval_fname, obs_fname, param_listing):
        return '''
            void ''' + fname + '''(const optimize_data* const data, mot_float_type* const x, mot_float_type* result){
                ''' + param_listing + '''
                double observation;
                double evaluation;
                for(int i = 0; i < ''' + str(inst_per_problem) + '''; i++){
                    observation = (double)''' + obs_fname + '''(data, i);
                    evaluation = (double)''' + eval_fname + '''(data, x, i);

                    result[i] = - ((evaluation * evaluation) / (2 * RicianNoise_sigma * RicianNoise_sigma))
                                + log_bessel_i0((observation * evaluation) / (RicianNoise_sigma * RicianNoise_sigma));
                }
            }
        '''

    def get_log_likelihood_function(self, fname, inst_per_problem, eval_fname, obs_fname, param_listing,
                                    full_likelihood=True):
        return '''
            double ''' + fname + '''(const optimize_data* const data, const mot_float_type* const x){
                ''' + param_listing + '''
                double sum = 0.0;
                double observation;
                double evaluation;
                for(int i = 0; i < ''' + str(inst_per_problem) + '''; i++){
                    observation = (double)''' + obs_fname + '''(data, i);
                    evaluation = (double)''' + eval_fname + '''(data, x, i);

                    sum += log(observation / (RicianNoise_sigma * RicianNoise_sigma))
                            - ((observation * observation) / (2 * RicianNoise_sigma * RicianNoise_sigma))
                            - ((evaluation * evaluation) / (2 * RicianNoise_sigma * RicianNoise_sigma))
                            + log_bessel_i0((observation * evaluation) / (RicianNoise_sigma * RicianNoise_sigma));
                }
                return sum;
            }
        '''

    def set_noise_level_std(self, noise_std):
        self.parameter_list[0].value = noise_std
        return self

    def get_noise_level_std(self):
        return self.parameter_list[0].value



import numpy as np
from ...cl_routines.base import CLRoutine


__author__ = 'Robbert Harms'
__date__ = "2014-02-05"
__license__ = "LGPL v3"
__maintainer__ = "Robbert Harms"
__email__ = "robbert.harms@maastrichtuniversity.nl"


class ErrorMeasures(CLRoutine):

    def __init__(self, cl_environments=None, load_balancer=None, double_precision=False):
        """Given a set of raw errors per voxel, calculate some interesting measures."""
        super(ErrorMeasures, self).__init__(cl_environments=cl_environments, load_balancer=load_balancer)
        self._double_precision = double_precision

    def calculate(self, errors):
        """Given a set of raw errors per voxel, calculate some interesting measures.

        Args:
            errors (ndarray): The list with errors per problem instance.

        Returns:
            A dictionary containing (for each voxel)
                - Errors.l2: the l2 norm (square root of sum of squares)
                - Errors.mse: the mean sum of squared errors
        """
        sse = np.sum(np.power(errors, 2), axis=1)
        return {'Errors.l2': np.linalg.norm(errors, axis=1),
                'Errors.mse': sse/errors.shape[1]}



import numpy as np

__author__ = 'Robbert Harms'
__date__ = "2016-10-03"
__maintainer__ = "Robbert Harms"
__email__ = "robbert.harms@maastrichtuniversity.nl"


class AbstractProblemData(object):
    """A simple data container for the data for optimization/sampling models."""

    @property
    def protocol(self):
        """Return the protocol data stored in this problem data container.

        The protocol data contains information about the experimental setup. In MRI this is the scanner protocol.

        Returns:
            dict: The protocol data dict.
        """
        return {}

    def get_nmr_inst_per_problem(self):
        """Get the number of instances/data points per problem.

        The minimum is one instance per problem.

        This number represents the number of data points

        Returns:
            int: the number of instances per problem.
        """
        return np.array(self.protocol[list(self.protocol.keys())[0]]).shape[0]

    def get_nmr_problems(self):
        """Get the number of problems present in this problem data.

        Returns:
            int: the number of problem instances
        """
        return self.observations.shape[0]

    @property
    def observations(self):
        """Return the observations stored in this problem data container.

        Returns:
            ndarray: The list of observed instances per problem. Should be a 2d matrix with as columns the observations
                and as rows the problems.
        """
        return np.array([[]])

    @property
    def static_maps(self):
        """Get a dictionary with the static maps.

        These maps will be loaded by the model builder as the values for the static parameters.

        Returns:
            Dict[str, value]: per static map the value for the static map. This can either be an one or two dimensional
                matrix containing the values for each problem instance or it can be a single value we will use
                for all p
        """
        return {}

    @property
    def noise_std(self):
        """The noise standard deviation we will use during model evaluation.

        During optimization or sampling the model will be evaluated against the observations using an evaluation
        model. Most of these evaluation models need to have a standard deviation.

        Returns:
            number of ndarray: either a scalar or a 2d matrix with one value per problem instance.
        """
        return 1


class SimpleProblemData(AbstractProblemData):

    def __init__(self, protocol, observations, static_maps=None, noise_std=None):
        """A simple data container for the data for optimization/sampling models.

        Args:
            protocol (dict): The protocol data dictionary
            observations (ndarray): The 2d array with the observations
            static_maps (dict): The dictionary with the static maps. These are 2d/3d ndarrays with one or more
                values per problem instance.
            noise_std (number or ndarray): either a scalar or a 2d matrix with one value per problem instance.
        """
        self._protocol = protocol
        self._observation = observations
        self._static_maps = static_maps or {}
        self._noise_std = noise_std

    @property
    def protocol(self):
        return self._protocol

    @property
    def observations(self):
        return self._observation

    @property
    def static_maps(self):
        return self._static_maps

    @property
    def noise_std(self):
        return self._noise_std



import numpy as np
from mot.data_adapters import SimpleDataAdapter
from mot.cl_data_type import CLDataType
from mot.model_building.cl_functions.parameters import CurrentObservationParam, StaticMapParameter, ProtocolParameter, \
    ModelDataParameter, FreeParameter
from mot.cl_routines.mapping.calc_dependent_params import CalculateDependentParameters
from mot.utils import TopologicalSort, is_scalar
from mot.model_building.parameter_functions.codecs import CodecBuilder
from mot.model_building.parameter_functions.dependencies import SimpleAssignment
from mot.model_interfaces import OptimizeModelInterface, SampleModelInterface

__author__ = 'Robbert Harms'
__date__ = "2014-03-14"
__license__ = "LGPL v3"
__maintainer__ = "Robbert Harms"
__email__ = "robbert.harms@maastrichtuniversity.nl"


class OptimizeModelBuilder(OptimizeModelInterface):

    def __init__(self, name, model_tree, evaluation_model, signal_noise_model=None, problem_data=None):
        """Create a new model builder that can construct an optimization model using parts.

        Args:
            name (str): the name of the model
            model_tree (CompartmentModelTree): the model tree object
            evaluation_model (EvaluationModel): the evaluation model to use for the resulting complete model
            signal_noise_model (SignalNoiseModel): the optional signal noise model to use to noise the model prediction
            problem_data (ProblemData): the problem data object

        Attributes;
            problems_to_analyze (list): the list with problems we want to analyze. Suppose we have a few thousands
                problems defined in this model, but we want to run the optimization only on a few problems. By setting
                this attribute to a list of problems indices only those problems will be analyzed.
        """
        super(OptimizeModelBuilder, self).__init__()
        self._name = name
        self._double_precision = False
        self._model_tree = model_tree
        self._evaluation_model = evaluation_model
        self._signal_noise_model = signal_noise_model
        self._parameters_dot_to_bar = {}
        self._dependency_store = DependencyStore()
        self._post_optimization_modifiers = []
        self.problems_to_analyze = None

        self._problem_data = None
        if problem_data:
            self.set_problem_data(problem_data)

        for m, p in self._get_model_parameter_list():
            self._parameters_dot_to_bar.update({m.name + '.' + p.name: m.name + '_' + p.name})

        self._check_for_double_model_names()
        self._set_default_dependencies()

    @property
    def name(self):
        """See super class OptimizeModelInterface for details"""
        return self._name

    @property
    def double_precision(self):
        """See super class OptimizeModelInterface for details"""
        return self._double_precision

    @double_precision.setter
    def double_precision(self, value):
        """Set the value for double_precision.

        Args:
            value (boolean): if we would like to do the computations in double of single floating point type.
        """
        self._double_precision = value

    def fix(self, model_param_name, value):
        """Fix the given model.param to the given value.

        Args:
            model_param_name (string): A model.param name like 'Ball.d'
            value (scalar or vector): The value to fix the given parameter to

        Returns:
            Returns self for chainability
        """
        m, p = self._get_model_parameter_matching(model_param_name)
        p.fix_to(value)
        return self

    def init(self, model_param_name, value):
        """Init the given model.param to the given value.

        Args:
            model_param_name (string): A model.param name like 'Ball.d'
            value (scalar or vector): The value to initialize the given parameter to

        Returns:
            Returns self for chainability
        """
        m, p = self._get_model_parameter_matching(model_param_name)
        p.value = value
        return self

    def set_lower_bound(self, model_param_name, value):
        """Set the lower bound for the given parameter to the given value.

        Args:
            model_param_name (string): A model.param name like 'Ball.d'
            value (scalar or vector): The value to set the lower bounds to

        Returns:
            Returns self for chainability
        """
        m, p = self._get_model_parameter_matching(model_param_name)
        p.lower_bound = value
        return self

    def set_upper_bound(self, model_param_name, value):
        """Set the upper bound for the given parameter to the given value.

        Args:
            model_param_name (string): A model.param name like 'Ball.d'
            value (scalar or vector): The value to set the upper bounds to

        Returns:
            Returns self for chainability
        """
        m, p = self._get_model_parameter_matching(model_param_name)
        p.upper_bound = value
        return self

    def set_parameter_transform(self, model_param_name, value):
        """Set the parameter transform for the given parameter to the given transformation.

        Args:
            model_param_name (string): A model.param name like 'Ball.d'
            value (AbstractTransformation): The parameter transform to use

        Returns:
            Returns self for chainability
        """
        m, p = self._get_model_parameter_matching(model_param_name)
        p.parameter_transform = value
        return self

    def unfix(self, model_param_name):
        """Unfix the given model.param

        Args:
            model_param_name (string): A model.param name like 'Ball.d'

        Returns:
            Returns self for chainability
        """
        m, p = self._get_model_parameter_matching(model_param_name)
        p.fixed = False
        return self

    def has_parameter(self, model_param_name):
        """Check to see if the given parameter is defined in this model.

        Args:
            model_param_name (string): A model.param name like 'Ball.d'

        Returns:
            boolean: true if the parameter is defined in this model, false otherwise.
        """
        for m, p in self._get_model_parameter_list():
            if m.name + '.' + p.name == model_param_name:
                return True
        return False

    def set_problem_data(self, problem_data):
        """Set the problem data this model will deal with.

        This will also call the function set_noise_level_std() with the noise_std from the new problem data.

        Args:
            problem_data (ProblemData):
                The container for the problem data we will use for this model.

        Returns:
            Returns self for chainability
        """
        self._problem_data = problem_data
        if self._problem_data.noise_std is not None:
            self._evaluation_model.set_noise_level_std(self._problem_data.noise_std)
        return self

    def cmf(self, model_name):
        """Get the Compartment Model Function object corresponding to the given model name.

        This may be useful for later fixing or adding items to a specific model.

        Args:
            model_name (str): the name of the compartment model to get

        Returns:
            CLFunction: the compartment model function with the given name. None if no matching function found.
        """
        models = self._get_model_list()
        for m in models:
            if m.name == model_name:
                return m
        return None

    def add_parameter_dependency(self, parameter_name, dependency):
        """Adds a dependency rule to this model. The dependency is supposed to be a ParameterDependency object.

        The dependencies are executed in the same order as they were added to this model.

        Args:
            parameter_name (String): The parameter on which the dependency is applied
            dependency (ParameterDependency): The dependency rule, an ParameterDependency object
        """
        if parameter_name not in self._parameters_dot_to_bar:
            raise ParameterNameException("The parameter name \"{}\" can not be "
                                         "found in the model listing.".format(parameter_name))
        self._dependency_store.set_dependency(parameter_name, dependency)
        return self

    def add_parameter_dependencies(self, dependencies):
        """Adds a list of dependency objects. The order of the dependencies matter (order of parameter instantiation).

        Args:
            dependencies: the dependency rules, an tuple list like ((name, ParameterDependency object),)
        """
        for name, d in dependencies:
            self.add_parameter_dependency(name, d)
        return self

    def add_post_optimization_modifier(self, model_param_name, mod_routine):
        """Add a modification function that can update the results of model optimization.

        The mod routine should be a function accepting a dictionary as input and should return a single map of
        the same dimension as the maps in the dictionary. The idea is that the mod_routine function gets the
        result dictionary from the optimization routine and calculates a new map.

        This map is returned and the dictionary is updated with the returned map as value and the here given
        model_param_name as key.

        It is possible to add more than one modifier function. In that case, they are called in the order they
        were appended to this model.

        Args:
            model_param_name (str): the parameter to which to add the modification routine
            mod_routine (python function): the callback function to apply on the results of the referenced parameter.
        """
        self._post_optimization_modifiers.append((model_param_name, mod_routine))
        return self

    def add_post_optimization_modifiers(self, modifiers):
        """Add a list of modifier functions.

        The same as add_post_optimization_modifier() except that it accepts a list of lists. Every element in the list
        should be a tuple like (model_param_name, mod_routine)

        Args:
            modifiers (tuple or list): The list of modifiers to add (in order).

        """
        self._post_optimization_modifiers.extend(modifiers)

    def get_required_protocol_names(self):
        """Get a list with the constant data names that are needed for this model to work.

        For example, an implementing diffusion MRI model might require the presence of the protocol parameter
        'g' and 'b'. This function should then return ('g', 'b').

        Returns:
            A list of columns names that are to be taken from the protocol data.
        """
        return list(set([p.name for m, p in self._get_model_parameter_list() if
                         isinstance(p, ProtocolParameter)]))

    def get_optimization_output_param_names(self):
        """See super class OptimizeModelInterface for details"""
        items = []
        for m in self._get_model_list():
            for p in m.get_free_parameters():
                items.append(m.name + '.' + p.name)

        for name, _ in self._post_optimization_modifiers:
            items.append(name)
        return items

    def get_optimized_param_names(self):
        return [m.name + '.' + p.name for m, p in self._get_estimable_parameters_list()]

    def get_nmr_problems(self):
        if self.problems_to_analyze is None:
            return self._problem_data.get_nmr_problems()
        return len(self.problems_to_analyze)

    def get_nmr_inst_per_problem(self):
        return self._problem_data.get_nmr_inst_per_problem()

    def get_nmr_estimable_parameters(self):
        return len(self.get_optimized_param_names())

    def get_problems_var_data(self):
        """See super class OptimizeModelInterface for details

        When overriding this function, please note that it should adhere to the attribute problems_to_analyze.
        """
        var_data_dict = {}

        observations = self._problem_data.observations
        if observations is not None:
            if self.problems_to_analyze is not None:
                observations = observations[self.problems_to_analyze, ...]

            observations = self._transform_observations(observations)

            data_adapter = SimpleDataAdapter(observations, CLDataType.from_string('mot_float_type*'),
                                             self._get_mot_float_type())
            var_data_dict.update({'observations': data_adapter})

        var_data_dict.update(self._get_fixed_parameters_as_var_data())
        var_data_dict.update(self._get_static_parameters_as_var_data())

        return var_data_dict

    def get_problems_protocol_data(self):
        protocol_info = self._problem_data.protocol
        return_data = {}
        for m, p in self._get_model_parameter_list():
            if isinstance(p, ProtocolParameter):
                if p.name in protocol_info:
                    if not self._all_elements_equal(protocol_info[p.name]):
                        const_d = {p.name: SimpleDataAdapter(protocol_info[p.name],
                                                             p.data_type, self._get_mot_float_type())}
                        return_data.update(const_d)
                else:
                    exception = 'Protocol parameter "{}" could not be resolved'.format(m.name + '.' + p.name)
                    raise ParameterResolutionException(exception)
        return return_data

    def get_model_data(self):
        model_data_dict = {}
        for m, p in self._get_model_parameter_list():
            if isinstance(p, ModelDataParameter) and not self._all_elements_equal(p.value):
                model_data_dict.update({p.name: SimpleDataAdapter(p.value, p.data_type, self._get_mot_float_type())})
        return model_data_dict

    def get_initial_parameters(self, results_dict=None):
        """When overriding this function, please note that it should adhere to the attribute problems_to_analyze.

        Args:
            results_dict (dict): the initialization settings for the specific parameters.
                The number of items per dictionary item should match the number of problems to analyze.
        """
        np_dtype = np.float32
        if self.double_precision:
            np_dtype = np.float64

        starting_points = []
        for m, p in self._get_estimable_parameters_list():
            if results_dict and (m.name + '.' + p.name) in results_dict:
                starting_points.append(results_dict[m.name + '.' + p.name])
            elif is_scalar(p.value):
                starting_points.append(np.full((self.get_nmr_problems(), 1), p.value, dtype=np_dtype))
            else:
                if len(p.value.shape) < 2:
                    value = np.transpose(np.asarray([p.value]))
                elif p.value.shape[1] > p.value.shape[0]:
                    value = np.transpose(p.value)
                else:
                    value = p.value

                if self.problems_to_analyze is None:
                    starting_points.append(value)
                else:
                    starting_points.append(value[self.problems_to_analyze, ...])

        starting_points = np.concatenate([np.transpose(np.array([s]))
                                          if len(s.shape) < 2 else s for s in starting_points], axis=1)

        data_adapter = SimpleDataAdapter(starting_points, CLDataType.from_string('mot_float_type'),
                                         self._get_mot_float_type())
        return data_adapter.get_opencl_data()

    def get_lower_bounds(self):
        return np.array([p.lower_bound for m, p in self._get_estimable_parameters_list()])

    def get_upper_bounds(self):
        return np.array([p.upper_bound for m, p in self._get_estimable_parameters_list()])

    def set_initial_parameters(self, initial_params):
        """Update the initial parameters for this model by the given values.

        This only affects free non fixed parameters.

        Args:
            initial_params (dict): a dictionary containing as keys full parameter names (<model>_<param>) and as values
                numbers or arrays to be used as starting point
        """
        for m, p in self._get_estimable_parameters_list():
            if m.name + '.' + p.name in initial_params:
                p.value = initial_params[m.name + '.' + p.name]
        return self

    def get_parameter_codec(self):
        dep_list = {}
        transform_dict = {}
        for m, p in self._get_estimable_parameters_list():
            name = m.name + '.' + p.name
            dep_names = list(dep[0].name + '.' + dep[1].name for dep in p.parameter_transform.dependencies)
            dep_list.update({name: dep_names})
            transform_dict.update({name: p})

        dep_list = TopologicalSort(dep_list).get_flattened()

        dec_func_list = []
        enc_func_list = []
        for name in dep_list:
            parameter = transform_dict[name]
            ind = self._get_parameter_estimable_index(name)
            transform = parameter.parameter_transform

            deps_names = []
            for dep in transform.dependencies:
                dep_ind = self._get_parameter_estimable_index(dep[0].name + '.' + dep[1].name)
                deps_names.append('{0}[' + str(dep_ind) + ']')

            s = '{0}[' + str(ind) + '] = ' + transform.get_cl_decode(parameter, '{0}[' + str(ind) + ']', deps_names)
            dec_func_list.append(s)

            s = '{0}[' + str(ind) + '] = ' + transform.get_cl_encode(parameter, '{0}[' + str(ind) + ']', deps_names)
            enc_func_list.append(s)

        return CodecBuilder(tuple(reversed(enc_func_list)), dec_func_list)

    def get_final_parameter_transformations(self, func_name='applyFinalParameterTransformations'):
        """Get the transformations that must be applied at the end of an optimization (or sampling) routine.

        These transformations must contain all parameter dependencies, as such that all transformation happening in the
        model function which do not happen in the codec must also go here.

        Args:
            func_name (str): the CL function name of the returned function

        Returns:
            str: A function of the kind:
                void finalParameterTransformations(const optimize_data* data, mot_float_type* x)
                Which is called for every voxel and must in place edit the x variable.
        """
        transform_needed = any(dp.has_side_effects or not dp.fixed for dp in
                               self._dependency_store.dependencies.values())

        if not self._dependency_store.has_dependencies() or not transform_needed:
            return None

        param_exclude_list = [m.name + '_' + p.name for (m, p) in self._get_non_model_tree_param_listing()]
        param_lists = self._get_parameter_type_lists()
        depend_param_listing = self._get_dependent_parameters_listing(param_lists['dependent'])

        for m, p in param_lists['fixed'] + param_lists['protocol']:
            if (m.name + '_' + p.name) not in depend_param_listing:
                param_exclude_list.append(m.name + '_' + p.name)

        param_listing = self._get_parameters_listing(exclude_list=param_exclude_list)

        func = "\n\t\t\t" + 'void ' + func_name + '(const optimize_data* const data, mot_float_type* const x){' + "\n"
        func += param_listing + "\n"

        for i, (m, p) in enumerate(self._get_parameter_type_lists()['estimable']):
            if not self._is_non_model_tree_model(m):
                func += "\t"*4 + 'x[' + str(i) + '] = ' + m.name + '_' + p.name + ';' + "\n"
        func += "\t\t\t" + '}' + "\n"
        return func

    def get_observation_return_function(self, func_name='getObservation'):
        if self._problem_data.observations.shape[1] < 2:
            return '''
                mot_float_type ''' + func_name + '''(const optimize_data* const data, const int observation_index){
                    return data->var_data_observations;
                }
            '''

        return '''
            mot_float_type ''' + func_name + '''(const optimize_data* const data, const int observation_index){
                return data->var_data_observations[observation_index];
            }
        '''

    def get_model_eval_function(self, func_name='evaluateModel'):
        noise_func_name = func_name + '_signalNoiseModel'
        func = self._get_model_functions_cl_code(noise_func_name)

        pre_model_function = self._get_pre_model_expression_eval_function()
        if pre_model_function:
            func += pre_model_function

        func += '''
            mot_float_type ''' + func_name + \
                '(const optimize_data* const data, const mot_float_type* const x, const int observation_index){' + "\n"

        func += self._get_parameters_listing(exclude_list=[m.name + '_' + p.name for (m, p) in
                                                           self._get_non_model_tree_param_listing()])

        if self._signal_noise_model:
            noise_params_listing = ''
            for p in self._signal_noise_model.get_free_parameters():
                noise_params_listing += "\t" * 4 + self._get_param_listing_for_param(self._signal_noise_model, p)
            func += "\n"
            func += noise_params_listing

        pre_model_code = self._get_pre_model_expression_eval_code()
        if pre_model_code:
            func += self._get_pre_model_expression_eval_code()

        func += "\n" + "\t"*4 + 'return ' + str(self._construct_model_expression(noise_func_name))
        func += "\n\t\t\t}"
        return func

    def get_objective_function(self, func_name="calculateObjective"):
        inst_per_problem = self.get_nmr_inst_per_problem()
        eval_func_name = func_name + '_evaluateModel'
        obs_func_name = func_name + '_getObservation'

        param_listing = ''
        for p in self._evaluation_model.get_free_parameters():
            param_listing += self._get_param_listing_for_param(self._evaluation_model, p)

        func = ''
        func += self._evaluation_model.get_cl_dependency_headers()
        func += self._evaluation_model.get_cl_dependency_code()

        func += self.get_model_eval_function(eval_func_name)
        func += self.get_observation_return_function(obs_func_name)
        func += str(self._evaluation_model.get_objective_function(func_name, inst_per_problem, eval_func_name,
                                                                  obs_func_name, param_listing))
        return str(func)

    def get_objective_list_function(self, func_name="calculateObjectiveList"):
        inst_per_problem = self.get_nmr_inst_per_problem()
        eval_func_name = func_name + '_evaluateModel'
        obs_func_name = func_name + '_getObservation'

        param_listing = ''
        for p in self._evaluation_model.get_free_parameters():
            param_listing += self._get_param_listing_for_param(self._evaluation_model, p)

        func = ''
        func += self._evaluation_model.get_cl_dependency_headers()
        func += self._evaluation_model.get_cl_dependency_code()

        func += self.get_model_eval_function(eval_func_name)
        func += self.get_observation_return_function(obs_func_name)
        func += str(self._evaluation_model.get_objective_list_function(func_name, inst_per_problem, eval_func_name,
                                                                       obs_func_name, param_listing))
        return str(func)

    def finalize_optimization_results(self, results_dict):
        """This adds the final optimization maps to the results dictionary.

        Steps in finalizing the results dict:
            1) It first adds the maps for the dependent and fixed parameters
            2) Second it adds the extra maps defined in the models itself.
            3) Third it loops through the post_optimization_modifiers callback functions for the final updates.
            4) Adds additional maps defined in this model subclass

        For more documentation see the base method.

        Args:
            results_dict (dict): the dictionary with the results (the keys are the parameter names, the values are the
                1d parameter lists)

        """
        self._add_dependent_parameter_maps(results_dict)
        self._add_fixed_parameter_maps(results_dict)

        for model in self._get_model_list():
            results_dict.update(model.get_extra_results_maps(results_dict))

        for name, routine in self._post_optimization_modifiers:
            results_dict[name] = routine(results_dict)

        self._add_finalizing_result_maps(results_dict)

        return results_dict

    def _add_fixed_parameter_maps(self, results_dict):
        """In place add complete maps for the fixed parameters."""
        param_lists = self._get_parameter_type_lists()
        fixed_params = param_lists['fixed']
        for (m, p) in fixed_params:
            if not self._parameter_fixed_to_dependency(m, p):
                name = m.name + '.' + p.name
                if is_scalar(p.value):
                    results_dict.update({name: np.tile(np.array([p.value]), (self.get_nmr_problems(),))})
                else:
                    value = p.value
                    if self.problems_to_analyze is not None:
                        value = value[self.problems_to_analyze, ...]
                    results_dict.update({name: value})

    def _add_dependent_parameter_maps(self, results_dict):
        """In place add complete maps for the dependent parameters."""
        param_lists = self._get_parameter_type_lists()
        if len(param_lists['dependent']):
            func = ''
            func += self._get_fixed_parameters_listing(param_lists['fixed'])
            func += self._get_estimable_parameters_listing(param_lists['estimable'])
            func += self._get_dependent_parameters_listing(param_lists['dependent'])

            estimable_params = [m.name + '.' + p.name for m, p in param_lists['estimable']]
            estimated_parameters = [results_dict[k] for k in estimable_params]

            dependent_parameter_names = [(m.name + '_' + p.name, m.name + '.' + p.name)
                                         for m, p in param_lists['dependent']]

            cpd = CalculateDependentParameters(double_precision=self.double_precision)
            dependent_parameters = cpd.calculate(self._get_fixed_parameters_as_var_data(),
                                                 estimated_parameters, func, dependent_parameter_names)

            results_dict.update(dependent_parameters)

    def _add_finalizing_result_maps(self, results_dict):
        """Add some final results maps to the results dictionary.

        This called by the function finalize_optimization_results() as last call to add more maps.

        Args:
            results_dict (args): the results from model optmization. We are to modify this in-place.
        """

    def _transform_observations(self, observations):
        """Apply a transformation on the observations before fitting.

        This function is called by get_problems_var_data() just before the observations are handed over to the
        CL routine, and just after the the list has been (optionally) limited with self.problems_to_analyze.

        To implement any behaviour here, you can override this function and add behaviour that changes the observations.

        Args:
            observations (ndarray): the 2d matrix with the observations. This is the list of
                observations *after* the list has been (optionally) limited with self.problems_to_analyze.

        Returns:
            observations (ndarray): a 2d matrix of the same shape as the input. This should hold the transformed data.
        """
        return observations

    def _construct_model_expression(self, noise_func_name):
        """Construct the model signel expression. This is supposed to be used in get_model_eval_function.

        Args:
            noise_func_name (str): the name of the noise function.
        """
        func = ''
        if self._signal_noise_model:
            noise_params_func = ''
            for p in self._signal_noise_model.get_free_parameters():
                noise_params_func += ', ' + self._signal_noise_model.name + '_' + p.name

            func += noise_func_name + '((' + self._build_model_from_tree(self._model_tree, 0) + ')' + \
                    noise_params_func + ');'
        else:
            func += '(' + self._build_model_from_tree(self._model_tree, 0) + ');'
        return func

    def _build_model_from_tree(self, node, depth):
        if not node.children:
            return self._model_to_string(node.data)
        else:
            subfuncs = []
            for child in node.children:
                if child.children:
                    subfuncs.append(self._build_model_from_tree(child, depth+1))
                else:
                    subfuncs.append(self._model_to_string(child.data))

            operator = node.data
            func = (' ' + operator + ' ').join(subfuncs)

        if func[0] == '(':
            return '(' + func + ')'
        return '(' + "\n" + ("\t" * int((depth/2)+5)) + func + "\n" + ("\t" * int((depth/2)+4)) + ')'

    def _model_to_string(self, model):
        """Convert a model to CL string."""
        param_list = []
        for param in model.parameter_list:
            if isinstance(param, ProtocolParameter):
                param_list.append(param.name)
            elif isinstance(param, ModelDataParameter):
                if self._all_elements_equal(param.value):
                    param_list.append(str(self._get_single_value(param.value)))
                else:
                    param_list.append('data->model_data_' + param.name)
            elif isinstance(param, StaticMapParameter):
                static_map_value = self._get_static_map_value(param)
                if self._all_elements_equal(static_map_value):
                    param_list.append(str(self._get_single_value(static_map_value)))
                else:
                    if len(static_map_value.shape) > 1 \
                            and static_map_value.shape[1] == self._problem_data.observations.shape[1]:
                        param_list.append('data->var_data_' + model.name + '_' + param.name + '[observation_index]')
                    else:
                        param_list.append('data->var_data_' + model.name + '_' + param.name)
            elif isinstance(param, CurrentObservationParam):
                param_list.append('data->var_data_observations[observation_index]')
            else:
                param_list.append(model.name + '_' + param.name)

        return model.cl_function_name + '(' + ', '.join(param_list) + ')'

    def _get_model_functions_cl_code(self, noise_func_name):
        """Get the model functions CL. This is used in get_model_eval_function()."""
        cl_code = ''
        for leave in self._model_tree.leaves:
            cl_code += leave.data.get_cl_header() + "\n"
            cl_code += leave.data.get_cl_code() + "\n"

        if self._signal_noise_model:
            cl_code += self._signal_noise_model.get_signal_function(noise_func_name)
        return cl_code

    def _get_parameters_listing(self, exclude_list=()):
        """Get the CL code for the parameter listing, this goes on top of the evaluate function.

        Args:
            exclude_list: an optional list containing parameters to exclude from the listing.
             This should contain full parameter names like: <model_name>_<param_name>

        Returns:
            An CL string that contains all the parameters as primitive data types.
        """
        func = ''
        param_lists = self._get_parameter_type_lists()
        func += self._get_protocol_parameters_listing(param_lists['protocol'], exclude_list=exclude_list)
        func += self._get_fixed_parameters_listing(param_lists['fixed'], exclude_list=exclude_list)
        func += self._get_estimable_parameters_listing(param_lists['estimable'], exclude_list=exclude_list)
        func += self._get_dependent_parameters_listing(param_lists['dependent'], exclude_list=exclude_list)
        return str(func)

    def _get_estimable_parameters_listing(self, param_list=None, exclude_list=()):
        """Get the parameter listing for the free parameters.

        For performance reasons, the parameter list should already be given.
            If not given it is calculated using:
                self._get_parameter_type_lists()['estimable']

        Args:
            param_list: the list with the estimable parameters
            exclude_list: a list of parameters to exclude from this listing
        """
        if param_list is None:
            param_list = self._get_parameter_type_lists()['estimable']

        func = ''
        estimable_param_counter = 0
        for m, p in param_list:
            name = m.name + '_' + p.name
            if name not in exclude_list:
                data_type = p.data_type.cl_type
                assignment = 'x[' + str(estimable_param_counter) + ']'
                func += "\t"*4 + data_type + ' ' + name + ' = ' + assignment + ';' + "\n"
                estimable_param_counter += 1
        return func

    def _get_protocol_parameters_listing(self, param_list=None, exclude_list=()):
        """Get the parameter listing for the protocol parameters.

        For performance reasons, the parameter list should already be given.
            If not given it is calculated using:
                self._get_parameter_type_lists()['protocol']

        Args:
            param_list: the list with the protocol parameters
            exclude_list: a list of parameters to exclude from this listing
        """
        protocol_info = self._problem_data.protocol
        if param_list is None:
            param_list = self._get_parameter_type_lists()['protocol']

        const_params_seen = []
        func = ''
        for m, p in param_list:
            if (m.name + '_' + p.name) not in exclude_list:
                data_type = p.data_type.cl_type
                if p.name not in const_params_seen:
                    if self._all_elements_equal(protocol_info[p.name]):
                        if p.data_type.is_vector_type:
                            vector_length = p.data_type.vector_length
                            values = [str(val) for val in protocol_info[p.name][0]]
                            if len(values) < vector_length:
                                values.append(str(0))
                            assignment = '(' + data_type + ')(' + ', '.join(values) + ')'
                        else:
                            assignment = str(float(protocol_info[p.name][0]))
                    else:
                        if p.data_type.is_pointer_type:
                            # this requires generic address spaces available in OpenCL >= 2.0.
                            assignment = '&data->protocol_data_' + p.name + '[observation_index]'
                        else:
                            assignment = 'data->protocol_data_' + p.name + '[observation_index]'
                    func += "\t"*4 + data_type + ' ' + p.name + ' = ' + assignment + ';' + "\n"
                    const_params_seen.append(p.name)
        return func

    def _get_fixed_parameters_listing(self, param_list=None, exclude_list=()):
        """Get the parameter listing for the fixed parameters.

        For performance reasons, the fixed parameter list should already be given.
            If not given it is calculated using:
                self._get_parameter_type_lists()['fixed']

        Args:
            dependent_param_list: the list list of fixed params
            exclude_list: a list of parameters to exclude from this listing
        """
        if param_list is None:
            param_list = self._get_parameter_type_lists()['fixed']

        func = ''
        for m, p in param_list:
            name = m.name + '_' + p.name
            if name not in exclude_list:
                data_type = p.data_type.raw_data_type
                if self._all_elements_equal(p.value):
                    assignment = '(' + data_type + ')' + str(float(self._get_single_value(p.value)))
                else:
                    assignment = '(' + data_type + ') data->var_data_' + m.name + '_' + p.name
                func += "\t"*4 + data_type + ' ' + name + ' = ' + assignment + ';' + "\n"
        return func

    def _get_dependent_parameters_listing(self, dependent_param_list=None, exclude_list=()):
        """Get the parameter listing for the dependent parameters.

        For performance reasons, the dependent parameter list should already be given.
            If not given it is calculated using:
                self._get_parameter_type_lists()['dependent']

        Args:
            dependent_param_list: the list list of dependent params
            exclude_list: a list of parameters to exclude from this listing, note that this will only exclude the
                definition of the parameter, not the dependency code.
        """
        if dependent_param_list is None:
            dependent_param_list = self._get_parameter_type_lists()['dependent']
        func = ''
        for m, p in dependent_param_list:
            pd = self._dependency_store.get_dependency(m.name + '.' + p.name)
            if pd.pre_transform_code:
                func += "\t"*4 + self._convert_parameters_dot_to_bar(pd.pre_transform_code)

            assignment = self._convert_parameters_dot_to_bar(pd.assignment_code)
            name = m.name + '_' + p.name
            data_type = p.data_type.raw_data_type

            if self._parameter_fixed_to_dependency(m, p):
                if (m.name + '_' + p.name) not in exclude_list:
                    func += "\t"*4 + data_type + ' ' + name + ' = ' + assignment + ';' + "\n"
            else:
                func += "\t"*4 + name + ' = ' + assignment + ';' + "\n"
        return func

    def _get_fixed_parameters_as_var_data(self):
        var_data_dict = {}
        for m, p in self._get_free_parameters_list():
            if p.fixed and not self._all_elements_equal(p.value) and not self._parameter_fixed_to_dependency(m, p):
                value = p.value
                if self.problems_to_analyze is not None:
                    value = value[self.problems_to_analyze, ...]

                var_data_dict.update({m.name + '_' + p.name: SimpleDataAdapter(value, p.data_type,
                                                                               self._get_mot_float_type())})
        return var_data_dict

    def _get_static_parameters_as_var_data(self):
        static_data_dict = {}

        for m, p in self._get_static_parameters_list():
            static_map_value = self._get_static_map_value(p)

            if not self._all_elements_equal(static_map_value):
                data_adapter = SimpleDataAdapter(static_map_value, p.data_type, self._get_mot_float_type())
                static_data_dict.update({m.name + '_' + p.name: data_adapter})

        return static_data_dict

    def _get_static_map_value(self, parameter):
        """Get the map value for the given parameter of the given model.

        This first checks if the parameter is defined in the static maps data in the problem data. If not, we try
        to get it from the value stored in the parameter itself. If that fails as well we raise an error.

        Also, this only returns the problems for which problems_to_analyze is set.

        Args:
            parameter (CLParameter): the parameter for which we want to get the value

        Returns:
            ndarray or number: the value for the given parameter.
        """
        data = None
        if parameter.name in self._problem_data.static_maps:
            data = self._problem_data.static_maps[parameter.name]
        elif parameter.value is not None:
            data = parameter.value

        if data is None:
            raise ValueError('No suitable data could be found for the static parameter {}.'.format(parameter.name))

        if is_scalar(data):
            return data

        if self.problems_to_analyze is not None:
            return data[self.problems_to_analyze, ...]
        return data

    def _get_non_model_tree_param_listing(self):
        listing = []
        for p in self._evaluation_model.parameter_list:
            listing.append((self._evaluation_model, p))

        if self._signal_noise_model:
            for p in self._signal_noise_model.parameter_list:
                listing.append((self._signal_noise_model, p))

        return listing

    def _is_non_model_tree_model(self, model):
        return model is self._evaluation_model or (self._signal_noise_model is not None and
                                                   model is self._signal_noise_model)

    def _get_param_listing_for_param(self, m, p):
        """Get the param listing for one specific parameter. This can be used for example for the noise model params.

        Please note, that on the moment this function does not support the complete dependency graph for the dependent
        parameters.
        """
        data_type = p.data_type.raw_data_type
        name = m.name + '_' + p.name
        assignment = ''

        if isinstance(p, ProtocolParameter):
            assignment = 'data->protocol_data_' + p.name + '[observation_index]'
        elif isinstance(p, FreeParameter):
            if p.fixed and not self._parameter_has_dependency(m, p):
                if self._all_elements_equal(p.value):
                    assignment = '(' + data_type + ')' + str(float(self._get_single_value(p.value)))
                else:
                    assignment = '(' + data_type + ') data->var_data_' + m.name + '_' + p.name
            elif not self._parameter_has_dependency(m, p) \
                or (self._parameter_has_dependency(m, p) and not self._parameter_fixed_to_dependency(m, p)):
                ind = self._get_parameter_estimable_index(m.name + '.' + p.name)
                assignment += 'x[' + str(ind) + ']'
            if self._parameter_has_dependency(m, p):
                return self._get_dependent_parameters_listing(((m, p),))

        return data_type + ' ' + name + ' = ' + assignment + ';' + "\n"

    def _get_parameter_type_lists(self):
        """Returns a dictionary with the parameters sorted in the types protocol, fixed, estimable and dependent.

        Parameters may occur in different lists (estimable and dependent for example).
        """
        protocol_parameters = []
        fixed_parameters = []
        estimable_parameters = []
        depended_parameters = []

        for m, p in self._get_model_parameter_list():
            if isinstance(p, ProtocolParameter):
                protocol_parameters.append((m, p))
            elif isinstance(p, FreeParameter):
                if p.fixed and not self._parameter_has_dependency(m, p):
                    fixed_parameters.append((m, p))
                elif not self._parameter_has_dependency(m, p) or (self._parameter_has_dependency(m, p)
                                                                  and not self._parameter_fixed_to_dependency(m, p)):
                    estimable_parameters.append((m, p))

                if self._parameter_has_dependency(m, p):
                    ind = self._dependency_store.get_index(m.name + '.' + p.name)
                    depended_parameters.insert(ind, (m, p))

        return {'protocol': protocol_parameters, 'fixed': fixed_parameters,
                'estimable': estimable_parameters, 'dependent': depended_parameters}

    def _get_parameter_by_name(self, parameter_name):
        """Get the parameter object of the given full parameter name in dot format.

        Args:
            parameter_name (string): the parameter name in dot format: <model>.<param>
        """
        models = self._get_model_list()
        for m in models:
            for p in m.parameter_list:
                if (m.name + '.' + p.name) == parameter_name:
                    return p

    def _get_parameter_estimable_index(self, parameter_name):
        """Get the index of this parameter in the parameters list

        This returns the position of this parameter in the 'x' vector in the CL kernel.

        Args:
            parameter_name: the parameter name in dot format. <model>.<param>
        """
        estimable_param_counter = 0
        for m in self._get_model_list():
            for p in m.parameter_list:
                if (m.name + '.' + p.name) == parameter_name:
                    return estimable_param_counter

                if isinstance(p, FreeParameter) \
                        and not self._parameter_fixed_to_dependency(m, p) \
                        and not p.fixed:
                    estimable_param_counter += 1

    def _get_model_list(self):
        """Get the list of all the Model Functions that play a role in this CompositeModel"""
        models = [n.data for n in self._model_tree.leaves]
        models.append(self._evaluation_model)
        if self._signal_noise_model:
            models.append(self._signal_noise_model)
        return models

    def _get_model_parameter_list(self):
        """Get a list of all model, parameter tuples."""
        return list((m, p) for m in self._get_model_list() for p in m.parameter_list)

    def _get_free_parameters_list(self):
        """Gets the free parameters (as model, parameter tuples) from the model listing.
        This does not incorporate checking for fixed parameters.
        """
        return list((m, p) for m in self._get_model_list() for p in m.get_free_parameters())

    def _get_static_parameters_list(self):
        """Gets the static parameters (as model, parameter tuples) from the model listing."""
        return list((m, p) for m in self._get_model_list() for p in m.get_parameters_of_type(StaticMapParameter))

    def _get_estimable_parameters_list(self):
        """Gets a list (as model, parameter tuples) of all parameters that are estimable. """
        l = []
        for m in self._get_model_list():
            for p in m.get_free_parameters():
                if self._parameter_estimable(m, p):
                    l.append((m, p))
        return l

    def _get_model_estimable_parameters(self, model):
        return list(p for p in model.get_free_parameters() if self._parameter_estimable(model, p))

    def _parameter_estimable(self, m, p):
        return not p.fixed and not self._parameter_fixed_to_dependency(m, p)

    def _parameter_fixed_to_dependency(self, model, param):
        return self._parameter_has_dependency(model, param) \
            and self._dependency_store.get_dependency(model.name + '.' + param.name).fixed

    def _parameter_has_dependency(self, model, param):
        """Check if the given model and parameter name combo has a dependency."""
        return self._dependency_store.has_dependency(model.name + '.' + param.name)

    def _convert_parameters_dot_to_bar(self, string):
        """Convert a string containing parameters with . to parameter names with _"""
        for dname, bname in self._parameters_dot_to_bar.items():
            string = string.replace(dname, bname)
        return string

    def _init_fixed_duplicates_dependencies(self):
        """Find duplicate fixed parameters, and make dependencies of them. This saves data transfer in CL."""
        var_data_dict = {}
        for m, p in self._get_free_parameters_list():
            if p.fixed \
                    and not is_scalar(p.value) \
                    and not self._parameter_fixed_to_dependency(m, p):

                duplicate_found = False
                duplicate_key = None

                for key, data in var_data_dict.items():
                    if np.array_equal(data, p.value):
                        duplicate_found = True
                        duplicate_key = key
                        break

                if duplicate_found:
                    self.add_parameter_dependency(m.name + '.' + p.name, SimpleAssignment(duplicate_key))
                else:
                    var_data_dict.update({m.name + '.' + p.name: p.value})

    def _check_for_double_model_names(self):
        models = self._get_model_list()
        model_names = []
        for m in models:
            if m.name in model_names:
                raise DoubleModelNameException("Double model name detected in the model tree.", m.name)
            model_names.append(m.name)

    def _contains_parameter_reference(self, s):
        """Test if the given string contains a parameter reference to another model.
        Here s is a string given by the user for a dependency, transformation or other possible string items.

        Args:
            s: string
                The string we need to check for possible user referenced parameters

        Returns:
            Tuple with tuples for each (model, param) referenced in the given string
        """
        return list((m, p) for m, p in self._get_model_parameter_list() if m.name + '.' + p.name in s)

    def _get_model_parameter_matching(self, model_param_name):
        for m, p in self._get_model_parameter_list():
            if m.name + '.' + p.name == model_param_name:
                return m, p
        raise ValueError('The parameter with the given name ({}) could not be found.'.format(model_param_name))

    def _get_pre_model_expression_eval_code(self):
        """The code called in the evaluation function.

        This is called after the parameters are initialized and before the model signal expression. It can call
        functions defined in _get_pre_model_expression_eval_function()

        Returns:
            str: cl code containing evaluation changes,
        """
        return ''

    def _get_pre_model_expression_eval_function(self):
        """Function used in the model evaluation generation function.

        The idea is that some implementing models may need to change some of the protocol or fixed parameters
        before they are handed over to the signal expression function. This function is called by the
        get_model_eval_function function during model evaluation function construction.

        Returns:
            str: cl function to be used in conjunction with the output of the function
                _get_pre_model_expression_eval_model()
        """

    def _set_default_dependencies(self):
        """Initialize the default dependencies.

        By default this only adds dependencies for the fixed data that is used in multiple parameters.
        """
        self._init_fixed_duplicates_dependencies()

    def _all_elements_equal(self, value):
        """Checks if all elements in the given value are equal to each other.

        If the input is a single value the result is trivial. Else we compare all the values to see
        if they are exactly the same.

        Args:
            value (ndarray or number): a numpy array or a single number.

        Returns:
            bool: true if all elements are equal to each other, false otherwise
        """
        if is_scalar(value):
            return True
        return (value == value[0]).all()

    def _get_single_value(self, value):
        """Get a single value out of the given value.

        This is meant to be used after a call to _all_elements_equal that returned True. With this
        function we return a single number from the input value.

        Args:
            value (ndarray or number): a numpy array or a single number.

        Returns:
            number: a single number from the input
        """
        if is_scalar(value):
            return value
        return value.item(0)

    def _get_mot_float_type(self):
        """Get the data type for the mot_float_type"""
        if self.double_precision:
            return CLDataType.from_string('double')
        return CLDataType.from_string('float')


class SampleModelBuilder(OptimizeModelBuilder, SampleModelInterface):

    def __init__(self, model_name, model_tree, evaluation_model, signal_noise_model=None, problem_data=None):
        super(SampleModelBuilder, self).__init__(model_name, model_tree, evaluation_model, signal_noise_model,
                                                 problem_data)

    def get_log_prior_function(self, func_name='getLogPrior'):
        prior = 'mot_float_type ' + func_name + '(const mot_float_type* const x){' + "\n"
        prior += "\t" + 'mot_float_type prior = 1.0;' + "\n"
        for i, (m, p) in enumerate(self._get_estimable_parameters_list()):
            prior += "\t" + 'prior *= ' + p.sampling_prior.get_cl_assignment(p, 'x[' + str(i) + ']') + "\n"
        prior += "\n" + "\t" + 'return log(prior);' + "\n" + '}'
        return prior

    def get_proposal_state(self):
        return_list = []
        for m, p in self._get_estimable_parameters_list():
            for param in p.sampling_proposal.get_parameters():
                if param.adaptable:
                    return_list.append(param.default_value)
        return return_list

    def is_proposal_symmetric(self):
        return all(p.sampling_proposal.is_symmetric for m, p in self._get_estimable_parameters_list())

    def get_proposal_logpdf(self, func_name='getProposalLogPDF'):
        return_str = ''
        for _, p in self._get_estimable_parameters_list():
            return_str += p.sampling_proposal.get_proposal_logpdf_function()

        return_str += "\n" + 'mot_float_type ' + func_name + \
            '(const int i, const mot_float_type proposal, const mot_float_type current, ' \
            ' mot_float_type* const proposal_state){' + "\n\t"

        return_str += "\n\t" + 'switch(i){' + "\n\t\t"

        adaptable_parameter_count = 0
        for i, (m, p) in enumerate(self._get_estimable_parameters_list()):
            return_str += 'case ' + str(i) + ':' + "\n\t\t\t"

            param_proposal = p.sampling_proposal
            logpdf_call = 'return ' + param_proposal.get_proposal_logpdf_function_name() + '(proposal, current'

            for param in param_proposal.get_parameters():
                if param.adaptable:
                    logpdf_call += ', proposal_state[' + str(adaptable_parameter_count) + ']'
                    adaptable_parameter_count += 1
                else:
                    logpdf_call += ', ' + str(param.default_value)

            logpdf_call += ');'
            return_str += logpdf_call + "\n"

        return_str += "\n\t\t" + '}' + "\n" + 'return 0;' + "\n"
        return_str += '}'
        return return_str

    def get_proposal_function(self, func_name='getProposal'):
        return_str = ''
        for _, p in self._get_estimable_parameters_list():
            return_str += p.sampling_proposal.get_proposal_function()

        return_str += "\n" + 'mot_float_type ' + func_name + \
            '(const int i, const mot_float_type current, void* rng_data, ' \
            ' mot_float_type* const proposal_state){'

        return_str += "\n\t" + 'switch(i){' + "\n\t\t"

        adaptable_parameter_count = 0
        for i, (m, p) in enumerate(self._get_estimable_parameters_list()):
            return_str += 'case ' + str(i) + ':' + "\n\t\t\t"

            param_proposal = p.sampling_proposal
            proposal_call = 'return ' + param_proposal.get_proposal_function_name() + '(current, rng_data'

            for param in param_proposal.get_parameters():
                if param.adaptable:
                    proposal_call += ', proposal_state[' + str(adaptable_parameter_count) + ']'
                    adaptable_parameter_count += 1
                else:
                    proposal_call += ', ' + str(param.default_value)

            proposal_call += ');'
            return_str += proposal_call + "\n"

        return_str += "\n\t\t" + '}' + "\n" + 'return 0;' + "\n"
        return_str += '}'
        return return_str

    def get_proposal_state_update_function(self, func_name='updateProposalParameters'):
        return_str = ''
        for _, p in self._get_estimable_parameters_list():
            for param in p.sampling_proposal.get_parameters():
                if param.adaptable:
                    return_str += param.get_parameter_update_function()

        return_str += 'void ' + func_name + '(uint* const ac_between_proposal_updates, ' + \
            'const uint proposal_update_intervals, mot_float_type* const proposal_state){' + "\n"

        adaptable_parameter_count = 0
        for i, (m, p) in enumerate(self._get_estimable_parameters_list()):
            param_proposal = p.sampling_proposal

            for param in param_proposal.get_parameters():
                if param.adaptable:
                    return_str += "\t" * 3
                    return_str += 'proposal_state[' + str(adaptable_parameter_count) + '] = '
                    return_str += param.get_parameter_update_function_name() + '(' +\
                        'proposal_state[' + str(adaptable_parameter_count) + '], ' + \
                        'ac_between_proposal_updates[' + str(i) + '], proposal_update_intervals);' + "\n"
                    adaptable_parameter_count += 1

        return_str += '}'
        return return_str

    def get_log_likelihood_function(self, func_name='getLogLikelihood', evaluation_model=None, full_likelihood=True):
        evaluation_model = evaluation_model or self._evaluation_model

        inst_per_problem = self.get_nmr_inst_per_problem()
        eval_func_name = func_name + '_evaluateModel'
        obs_func_name = func_name + '_getObservation'

        param_listing = ''
        for p in evaluation_model.get_free_parameters():
            param_listing += self._get_param_listing_for_param(evaluation_model, p)

        func = ''
        func += evaluation_model.get_cl_dependency_headers()
        func += evaluation_model.get_cl_dependency_code()

        func += self.get_model_eval_function(eval_func_name)
        func += self.get_observation_return_function(obs_func_name)
        func += evaluation_model.get_log_likelihood_function(func_name, inst_per_problem, eval_func_name,
                                                             obs_func_name, param_listing,
                                                             full_likelihood=full_likelihood)
        return func

    def samples_to_statistics(self, samples_dict):
        results = {}
        for key, value in samples_dict.items():
            param = self._get_parameter_by_name(key)
            stat_mod = param.sampling_statistics
            results[key] = stat_mod.get_mean(value)
            results[key + '.std'] = stat_mod.get_std(value)
        return results


class DependencyStore(object):

    def __init__(self):
        self.names_in_order = []
        self.dependencies = {}

    def set_dependency(self, param_name, dependency):
        if param_name not in self.names_in_order:
            self.names_in_order.append(param_name)
        self.dependencies.update({param_name: dependency})

    def get_dependency(self, param_name):
        return self.dependencies[param_name]

    def has_dependency(self, param_name):
        return param_name in self.dependencies

    def has_dependencies(self):
        return self.names_in_order

    def get_index(self, param_name):
        return self.names_in_order.index(param_name)


class ParameterNameException(Exception):
    """Thrown when the a parameter of an given name could not be found."""
    pass


class ParameterResolutionException(Exception):
    """Thrown when a fixed parameter could not be resolved."""
    pass


class DoubleModelNameException(Exception):
    """Thrown when there are two models with the same name."""
    pass



import logging
import numpy as np
import pyopencl as cl
from mot.cl_routines.mapping.error_measures import ErrorMeasures
from mot.cl_routines.mapping.residual_calculator import ResidualCalculator
from ...utils import results_to_dict, ParameterCLCodeGenerator, get_float_type_def
from ...cl_routines.base import CLRoutine
from ...load_balance_strategies import Worker
from ...cl_routines.mapping.final_parameters_transformer import FinalParametersTransformer
from ...cl_routines.mapping.codec_runner import CodecRunner
from ...__version__ import __version__

__author__ = 'Robbert Harms'
__date__ = "2014-05-18"
__license__ = "LGPL v3"
__maintainer__ = "Robbert Harms"
__email__ = "robbert.harms@maastrichtuniversity.nl"


return_code_labels = {
    0: ['default', 'no return code specified'],
    1: ['found zero', 'sum of squares/evaluation below underflow limit'],
    2: ['converged', 'the relative error in the sum of squares/evaluation is at most tol'],
    3: ['converged', 'the relative error of the parameter vector is at most tol'],
    4: ['converged', 'both errors are at most tol'],
    5: ['trapped', 'by degeneracy; increasing epsilon might help'],
    6: ['exhausted', 'number of function calls exceeding preset patience'],
    7: ['failed', 'ftol<tol: cannot reduce sum of squares any further'],
    8: ['failed', 'xtol<tol: cannot improve approximate solution any further'],
    9: ['failed', 'gtol<tol: cannot improve approximate solution any further'],
    10: ['NaN', 'Function value is not-a-number or infinite'],
    11: ['exhausted', 'temperature decreased to 0.0']
}


class AbstractOptimizer(CLRoutine):

    def __init__(self, cl_environments=None, load_balancer=None, use_param_codec=True, patience=1,
                 optimizer_settings=None, **kwargs):
        """Create a new optimizer that will minimize the given model with the given codec using the given environments.

        If the codec is None it is not used, if the environment is None, a suitable default environment should be
        created.

        Args:
            cl_environments (list of CLEnvironment): a list with the cl environments to use
            load_balancer (LoadBalanceStrategy): the load balance strategy to use
            use_param_codec (boolean): if this minimization should use the parameter codecs (param transformations)
            patience (int): The patience is used in the calculation of how many iterations to iterate the optimizer.
                The exact usage of this value of this parameter may change per optimizer.
            optimizer_options (dict): extra options one can set for the optimization routine. These are routine
                dependent.
        """
        self._optimizer_settings = optimizer_settings or {}
        if not isinstance(self._optimizer_settings, dict):
            raise ValueError('The given optimizer settings is not a dictionary.')

        self._use_param_codec = use_param_codec
        self.patience = patience or 1

        if 'patience' in self._optimizer_settings:
            self.patience = self._optimizer_settings['patience']
        self._optimizer_settings['patience'] = self.patience

        super(AbstractOptimizer, self).__init__(cl_environments, load_balancer, **kwargs)

    @property
    def use_param_codec(self):
        """Check if we will use the codec during optimization

        Returns:
            boolean: True if we will use the codec, false otherwise.
        """
        return self._use_param_codec

    @use_param_codec.setter
    def use_param_codec(self, use_param_codec):
        """Set if we will use the codec during optimization

        Args:
            use_param_codec (boolean): Set the value of use_param_codec.
        """
        self._use_param_codec = use_param_codec

    @property
    def optimizer_settings(self):
        """Get the optimization flags and settings set to the optimizer.

        This ordinarily should contain all the extra flags and options set to the optimization routine. Even those
        set additionally in the constructor.

        Returns:
            dict: the optimization options (settings)
        """
        return self._optimizer_settings

    def minimize(self, model, init_params=None, full_output=False):
        """Minimize the given model with the given codec using the given environments.

        Args:
            model (AbstractModel): The model to minimize, instance of AbstractModel
            init_params (dict): A dictionary containing the results of a previous run, provides the starting point
            full_output (boolean): If set to true, the output is a list with the results and a dictionary
                with other outputs. If false, only return the results per problem instance.

        Returns:
            Either only the results per problem, or a list: (results, {}), depending on if full_output is set.
            Both the results dictionary as well as the extra output  dictionary should return results per
            problem instance.
        """


class AbstractParallelOptimizer(AbstractOptimizer):

    def __init__(self, **kwargs):
        """
        Args:
            optimizer_settings (dict): extra options one can set for the optimization routine. These are routine
                dependent.
        """
        super(AbstractParallelOptimizer, self).__init__(**kwargs)
        self._logger = logging.getLogger(__name__)

    def minimize(self, model, init_params=None, full_output=False):
        self._logger.info('Entered optimization routine.')
        self._logger.info('Using MOT version {}'.format(__version__))
        self._logger.info('We will use a {} precision float type for the calculations.'.format(
            'double' if model.double_precision else 'single'))
        for env in self.load_balancer.get_used_cl_environments(self.cl_environments):
            self._logger.info('Using device \'{}\'.'.format(str(env)))
        self._logger.debug('Using compile flags: {}'.format(self.get_compile_flags_list()))
        self._logger.info('The parameters we will optimize are: {0}'.format(model.get_optimized_param_names()))
        self._logger.info('We will use the optimizer {} '
                          'with optimizer settings {}'.format(self.__class__.__name__,
                                                                             self.patience,
                                                                             self._optimizer_settings))

        self._logger.info('Starting optimization preliminaries')

        np_dtype = np.float32
        if model.double_precision:
            np_dtype = np.float64

        starting_points = np.require(model.get_initial_parameters(init_params), np_dtype,
                                     requirements=['C', 'A', 'O', 'W'])
        nmr_params = starting_points.shape[1]

        var_data_dict = model.get_problems_var_data()
        protocol_data_dict = model.get_problems_protocol_data()
        model_data_dict = model.get_model_data()
        return_codes = np.zeros((starting_points.shape[0],), dtype=np.int8, order='C')

        space_transformer = CodecRunner(self.cl_environments, self.load_balancer, model.double_precision)
        param_codec = model.get_parameter_codec()
        if self.use_param_codec and param_codec:
            starting_points = space_transformer.encode(param_codec, starting_points)
            starting_points = np.require(starting_points, np_dtype, requirements=['C', 'A', 'O', 'W'])

        self._logger.info('Finished optimization preliminaries')
        self._logger.info('Starting optimization')

        workers = self._create_workers(self._get_worker_generator(self, model, starting_points,
                                                                  var_data_dict, protocol_data_dict, model_data_dict,
                                                                  nmr_params, return_codes,
                                                                  self._optimizer_settings))
        self.load_balancer.process(workers, model.get_nmr_problems())
        del workers

        self._logger.info('Finished optimization')

        self._logger.info('Starting post-optimization transformations')
        results = FinalParametersTransformer(cl_environments=self._cl_environments,
                                             load_balancer=self.load_balancer).transform(model, starting_points)
        self._logger.info('Finished post-optimization transformations')

        self._logger.info('Calling finalize optimization results in the model')
        results = model.finalize_optimization_results(results_to_dict(results, model.get_optimized_param_names()))

        self._logger.info('Optimization finished.')

        if full_output:
            extra_output = {'ReturnCodes': return_codes}
            self._logger.info('Calculating errors measures')
            errors = ResidualCalculator(cl_environments=self.cl_environments, load_balancer=self.load_balancer).\
                calculate(model, results)
            extra_output.update(ErrorMeasures(self.cl_environments, self.load_balancer,
                                              model.double_precision).calculate(errors))
            self._logger.info('Done calculating errors measures')

            return results, extra_output
        return results

    def _get_worker_generator(self, *args):
        """Generate the worker generator callback for the function _create_workers()

        This is supposed to be overwritten by the implementing optimizer.

        Returns:
            the python callback for generating the worker
        """
        return lambda cl_environment: AbstractParallelOptimizerWorker(cl_environment, *args)


class AbstractParallelOptimizerWorker(Worker):

    def __init__(self, cl_environment, parent_optimizer, model, starting_points,
                 var_data_dict, protocol_data_dict, model_data_dict, nmr_params, return_codes,
                 optimizer_settings=None):
        super(AbstractParallelOptimizerWorker, self).__init__(cl_environment)

        self._optimizer_settings = optimizer_settings

        self._parent_optimizer = parent_optimizer

        self._model = model
        self._double_precision = model.double_precision
        self._nmr_params = nmr_params
        self._var_data_dict = var_data_dict
        self._protocol_data_dict = protocol_data_dict
        self._model_data_dict = model_data_dict
        self._return_codes = return_codes

        param_codec = model.get_parameter_codec()
        self._use_param_codec = self._parent_optimizer.use_param_codec and param_codec

        self._starting_points = starting_points
        self._all_buffers, self._params_buffer, self._return_code_buffer = self._create_buffers()
        self._kernel = self._build_kernel(self._parent_optimizer.get_compile_flags_list())

    def __del__(self):
        list(buffer.release() for buffer in self._all_buffers)

    def calculate(self, range_start, range_end):
        nmr_problems = range_end - range_start

        kernel_event = self._kernel.minimize(self._cl_run_context.queue, (nmr_problems, ), None, *self._all_buffers,
                                             global_offset=(range_start,))
        return [
            self._enqueue_readout(self._params_buffer, self._starting_points, range_start, range_end, [kernel_event]),
            self._enqueue_readout(self._return_code_buffer, self._return_codes, range_start, range_end, [kernel_event])
        ]

    def _create_buffers(self):
        all_buffers = []

        parameters_buffer = cl.Buffer(self._cl_run_context.context,
                                      cl.mem_flags.READ_WRITE | cl.mem_flags.USE_HOST_PTR,
                                      hostbuf=self._starting_points)
        all_buffers.append(parameters_buffer)

        return_code_buffer = cl.Buffer(self._cl_run_context.context,
                                       cl.mem_flags.WRITE_ONLY | cl.mem_flags.USE_HOST_PTR,
                                       hostbuf=self._return_codes)
        all_buffers.append(return_code_buffer)

        for data in self._var_data_dict.values():
            all_buffers.append(cl.Buffer(self._cl_run_context.context,
                                         cl.mem_flags.READ_ONLY | cl.mem_flags.COPY_HOST_PTR,
                                         hostbuf=data.get_opencl_data()))

        all_buffers.extend(self._generate_constant_buffers(self._protocol_data_dict, self._model_data_dict))

        return all_buffers, parameters_buffer, return_code_buffer

    def _get_kernel_source(self):
        """Generate the kernel source for this optimization routine.

        By default this returns a full kernel source using information from _get_optimizer_cl_code()
        and _get_optimizer_call_name().

        One could overwrite this function to completely generate the kernel source, but most likely
        you would want to implement _get_optimizer_cl_code() and _get_optimizer_call_name().

        Returns:
            str: The kernel source for this optimization routine.
        """
        nmr_params = self._nmr_params
        param_code_gen = ParameterCLCodeGenerator(self._cl_environment.device,
                                                  self._var_data_dict,
                                                  self._protocol_data_dict,
                                                  self._model_data_dict)

        kernel_source = ''
        kernel_source += get_float_type_def(self._double_precision)
        kernel_source += str(param_code_gen.get_data_struct())

        if self._use_param_codec:
            param_codec = self._model.get_parameter_codec()
            decode_func = param_codec.get_cl_decode_function('decodeParameters')
            kernel_source += decode_func + "\n"

        kernel_source += self._get_optimizer_cl_code()
        kernel_source += '''
            __kernel void minimize(
                ''' + ",\n".join(self._get_kernel_param_names(param_code_gen)) + '''
                ){
                    int gid = get_global_id(0);
        '''

        kernel_source += '''
                    mot_float_type x[''' + str(nmr_params) + '''];
                    for(int i = 0; i < ''' + str(nmr_params) + '''; i++){
                        x[i] = params[gid * ''' + str(nmr_params) + ''' + i];
                    }

                    ''' + param_code_gen.get_data_struct_init_assignment('data') + '''
                    return_codes[gid] = (char) ''' + self._get_optimizer_call_name() + '''(''' + ', '.join(self._get_optimizer_call_args()) + ''');

                    ''' + ('decodeParameters(x);' if self._use_param_codec else '') + '''

                    for(int i = 0; i < ''' + str(nmr_params) + '''; i++){
                        params[gid * ''' + str(nmr_params) + ''' + i] = x[i];
                    }
                }
        '''
        return kernel_source

    def _get_kernel_param_names(self, param_code_gen):
        """Get the list of kernel parameter names.

        This is useful if a subclass extended the buffers with an additional buffer.

        Args:
            param_code_gen (ParameterCLCodeGenerator): the kernel parameter code generator for the model kernel params.

        Returns:
            list of str: the list of kernel parameter names
        """
        kernel_param_names = ['global mot_float_type* params',
                              'global char* return_codes']
        kernel_param_names.extend(param_code_gen.get_kernel_param_names())

        if self._uses_random_numbers():
            kernel_param_names.append('global ranluxcl_state_t* ranluxcltab')

        return kernel_param_names

    def _get_optimizer_call_args(self):
        """Get the optimizer calling arguments.

        This is useful if a subclass extended the buffers with an additional buffer.

        Returns:
            list of str: the list of optimizer call arguments
        """
        call_args = ['x', '(const void*) &data']
        if self._uses_random_numbers():
            call_args.append('(void*) &ranluxclstate')
        return call_args

    def _get_optimizer_cl_code(self):
        """Get the optimization CL code that is called during optimization for each voxel.

        This is normally called by the default implementation of _get_kernel_source().

        By default this creates a CL function named 'evaluation' that can be called by the optimization routine.
        This default function takes into account the use of the parameter codec and calls the objective function
        of the model to actually evaluate the model.

        Returns:
            str: The kernel source for the optimization routine.
        """
        kernel_source = self._get_evaluate_function()
        kernel_source += self._get_optimization_function()
        return kernel_source

    def _get_evaluate_function(self):
        """Get the CL code for the evaluation function. This is called from _get_optimizer_cl_code.

        Implementing optimizers can change this if desired.

        Returns:
            str: the evaluation function.
        """
        kernel_source = ''
        kernel_source += self._model.get_objective_function('calculateObjective')
        if self._use_param_codec:
            kernel_source += '''
                mot_float_type evaluate(mot_float_type* x, const void* data){
                    mot_float_type x_model[''' + str(self._nmr_params) + '''];
                    for(int i = 0; i < ''' + str(self._nmr_params) + '''; i++){
                        x_model[i] = x[i];
                    }
                    decodeParameters(x_model);
                    return calculateObjective((optimize_data*)data, x_model);
                }
            '''
        else:
            kernel_source += '''
                mot_float_type evaluate(mot_float_type* x, const void* data){
                    return calculateObjective((optimize_data*)data, x);
                }
            '''
        return kernel_source

    def _get_optimization_function(self):
        """Return the optimization function as a CL string for the implementing optimizer.

        This is a convenience function to avoid boilerplate in implementing _get_optimizer_cl_code().

        Returns:
            str: The optimization routine function
        """

    def _get_optimizer_call_name(self):
        """Get the call name of the optimization routine.

        This name is the name of the function called by the kernel to optimize a single voxel.

        Returns:
            str: The function name of the optimization function.
        """
        return ''


__author__ = 'Robbert Harms'
__date__ = "2014-06-20"
__license__ = "LGPL v3"
__maintainer__ = "Robbert Harms"
__email__ = "robbert.harms@maastrichtuniversity.nl"


class AbstractTransformation(object):

    def __init__(self, dependencies=()):
        """The transformations define the encode and decode operations needed to build a codec.

        These objects define the basic transformation from and to model and optimization space.

        The state of the parameters is extrinsic. Calling the encode and decode functions needs a reference to a
        parameter, these should be handled by the client code.

        The transformation function may depend on other parameters. (Client code should handle the correct order
        of handling the transformations given all the dependencies.)

        Args:
            dependencies (list of (CLFunction, CLFunctionParameter) pairs): A list of (models, parameters)
                which are necessary for the operation of the transformation.
        """
        super(AbstractTransformation, self).__init__()
        self._dependencies = dependencies

    def get_cl_encode(self, parameter, parameter_name, dependencies_names=()):
        """Get the CL encode string

        Args:
            parameter (CLFunctionParameter): The parameter used as context for building the encode function.
            parameter_name (str): The name of this parameter, needed to generate the cl code
            dependencies_names (list of str): For each dependency (ordered) give the name how it should be used in the
                cl code.

        Returns:
            str: The cl code that can encode the parameter.
        """
        pass

    def get_cl_decode(self, parameter, parameter_name, dependencies_names=()):
        """Get the CL decode string

        Args:
            parameter (CLFunctionParameter): The parameter used as context for building the decode function.
            parameter_name (str): The name of this parameter, needed to generate the cl code
            dependencies_names (list of str): For each dependency (ordered) give the name how it should be used in the
                cl code.

        Returns:
            str: The cl code that can decode the parameter.
        """
        pass

    @property
    def dependencies(self):
        """Get a list of (CLFunction, CLFunctionParameter) pairs where this transformation depends on

        Returns:
            A list of (CLFunction, CLFunctionParameter) pairs.
        """
        return self._dependencies


class SimpleTransformation(AbstractTransformation):
    def __init__(self, cl_encode, cl_decode):
        """Adds a simple parameter transformation rule.

        This is for one parameter, a simple one-line transformation transformation.

        Args:
            cl_to_encoded (str): the assignment that transforms the parameter from model space to encoded space.
                It should contain one {} used for the parameter_name variable.
            cl_to_model (str): the assignment that transforms the parameter from encoded space to model space.
                It should contain one {} used for the parameter_name variable.
        """
        super(SimpleTransformation, self).__init__()
        self._cl_encode = cl_encode
        self._cl_decode = cl_decode

    def get_cl_encode(self, parameter, parameter_name, dependencies_names=()):
        return self._cl_encode.format(parameter_name)

    def get_cl_decode(self, parameter, parameter_name, dependencies_names=()):
        return self._cl_decode.format(parameter_name)


class IdentityTransform(SimpleTransformation):
    def __init__(self):
        """The identity transform does no transformation and returns the input given."""
        super(IdentityTransform, self).__init__('{};', '{};')


class ClampTransform(AbstractTransformation):
    """The clamp transformation limits the parameter between its lower and upper bound using the clamp function."""

    def get_cl_encode(self, parameter, parameter_name, dependencies_names=()):
        return 'clamp((mot_float_type)' + parameter_name + ', (mot_float_type)' + str(parameter.lower_bound) + \
               ', (mot_float_type)' + str(parameter.upper_bound) + ');'

    def get_cl_decode(self, parameter, parameter_name, dependencies_names=()):
        return 'clamp((mot_float_type)' + parameter_name + ', (mot_float_type)' + str(parameter.lower_bound) + \
               ', (mot_float_type)' + str(parameter.upper_bound) + ');'


class CosSqrClampTransform(AbstractTransformation):
    """The clamp transformation limits the parameter between its lower and upper bound using a cos(sqr()) transform."""

    def get_cl_encode(self, parameter, parameter_name, dependencies_names=()):
        template = 'acos(clamp(sqrt(fabs( ({param_name} - {lower_bound}) / ({upper_bound} - {lower_bound}) )), ' \
                   '    (mot_float_type)-1, (mot_float_type)1));'
        return template.format(param_name=parameter_name, lower_bound=parameter.lower_bound,
                             upper_bound=parameter.upper_bound)

    def get_cl_decode(self, parameter, parameter_name, dependencies_names=()):
        return 'pown(cos(' + parameter_name + '), 2) * ' + \
               '(' + str(parameter.upper_bound) + ' - ' + str(parameter.lower_bound) + ')' + \
                ' + ' + str(parameter.lower_bound) + ';'


class SinSqrClampTransform(AbstractTransformation):
    """The clamp transformation limits the parameter between its lower and upper bound using a sin(sqr()) transform."""

    def get_cl_encode(self, parameter, parameter_name, dependencies_names=()):
        template = 'asin(clamp(sqrt(fabs( ({param_name} - {lower_bound}) / ({upper_bound} - {lower_bound}) )), ' \
                   '    (mot_float_type)-1, (mot_float_type)1));'
        return template.format(param_name=parameter_name, lower_bound=parameter.lower_bound,
                               upper_bound=parameter.upper_bound)

    def get_cl_decode(self, parameter, parameter_name, dependencies_names=()):
        return 'pown(sin(' + parameter_name + '), 2) * ' + \
               '(' + str(parameter.upper_bound) + ' - ' + str(parameter.lower_bound) + ')' + \
               ' + ' + str(parameter.lower_bound) + ';'


class SqrClampTransform(AbstractTransformation):
    """The clamp transformation limits the parameter between its lower and upper bound using a sqr() transform."""

    def get_cl_encode(self, parameter, parameter_name, dependencies_names=()):
        return 'sqrt(' + parameter_name + ');'

    def get_cl_decode(self, parameter, parameter_name, dependencies_names=()):
        return 'clamp((mot_float_type)(' + parameter_name + ' * ' + parameter_name + '), (mot_float_type)' + \
               str(parameter.lower_bound) + ', (mot_float_type)' + str(parameter.upper_bound) + ');'


class SinSqrClampDependentTransform(AbstractTransformation):
    """The clamp transformation limits the parameter between 0 and the given parameter with the sin(sqr()) transform."""

    def get_cl_encode(self, parameter, parameter_name, dependencies_names=()):
        return 'asin(sqrt(fabs((' + parameter_name + ' - ' + \
               str(parameter.lower_bound) + ') / (' + \
               dependencies_names[0] + ' - ' + \
               str(parameter.lower_bound) + '))));'

    def get_cl_decode(self, parameter, parameter_name, dependencies_names=()):
        return 'pown(sin(' + parameter_name + '), 2) * ' + dependencies_names[0] + ' + ' \
                    + str(parameter.lower_bound) + ';'


class AbsModXTransform(AbstractTransformation):
    def __init__(self, x, dependencies=()):
        """Create an transformation that returns the absolute modulo x value of the input."""
        super(AbsModXTransform, self).__init__(dependencies)
        self._x = x

    def get_cl_encode(self, parameter, parameter_name, dependencies_names=()):
        return 'fmod((mot_float_type)fabs(' + parameter_name + '), (mot_float_type)' + str(self._x) + ');'

    def get_cl_decode(self, parameter, parameter_name, dependencies_names=()):
        return 'fmod((mot_float_type)fabs(' + parameter_name + '), (mot_float_type)' + str(self._x) + ');'


class CosSqrTransform(SimpleTransformation):
    def __init__(self):
        super(CosSqrTransform, self).__init__('acos(sqrt(fabs({})));', 'pown(cos({}), 2);')


class SinSqrTransform(SimpleTransformation):
    def __init__(self):
        super(SinSqrTransform, self).__init__('asin(sqrt(fabs({})));', 'pown(sin({}), 2);')


class AbsModPiTransform(AbsModXTransform):
    def __init__(self):
        super(AbsModPiTransform, self).__init__('M_PI')


class SqrTransform(SimpleTransformation):
    def __init__(self):
        super(SqrTransform, self).__init__('sqrt(fabs({}));', 'pown({}, 2);')



import numpy as np

__author__ = 'Robbert Harms'
__date__ = "2014-10-23"
__license__ = "LGPL v3"
__maintainer__ = "Robbert Harms"
__email__ = "robbert.harms@maastrichtuniversity.nl"


class ParameterSampleStatistics(object):
    def get_mean(self, samples):
        """Given the distribution represented by this statistic, get the mean of the samples.

        Args:
            samples (ndarray): The 2d array with the samples per voxel.

        Returns:
            A 1d ndarray with the mean per voxel.
        """

    def get_std(self, samples):
        """Given the distribution represented by this statistic, get the standard deviation of the samples.

        Args:
            samples (ndarray): The 2d array with the samples per voxel.

        Returns:
            A 1d array with the variance per voxel.
        """


class GaussianPSS(ParameterSampleStatistics):
    def get_mean(self, samples):
        return np.mean(samples, axis=1)

    def get_std(self, samples):
        return np.std(samples, axis=1)


class CircularGaussianPSS(ParameterSampleStatistics):
    def __init__(self, max_angle=np.pi):
        """Compute the circular mean for samples in a range

        The minimum angle is set to 0, the maximum angle can be given.

        Args:
            max_angle (number): The maximum angle used in the calculations
        """
        super(CircularGaussianPSS, self).__init__()
        self.max_angle = max_angle

    def get_mean(self, samples):
        return CircularGaussianPSS.circmean(np.mod(samples, self.max_angle), high=self.max_angle, low=0, axis=1)

    def get_std(self, samples):
        return CircularGaussianPSS.circstd(np.mod(samples, self.max_angle), high=self.max_angle, low=0, axis=1)

    @staticmethod
    def circmean(samples, high=2 * np.pi, low=0, axis=None):
        """Compute the circular mean for samples in a range.
        Taken from scipy.stats

        Args:
            samples (array_like): Input array.
            high (float or int): High boundary for circular mean range.  Default is ``2*pi``.
            low (float or int): Low boundary for circular mean range.  Default is 0.
            axis (int, optional): Axis along which means are computed.
                The default is to compute the mean of the flattened array.

        Returns:
            float: Circular mean.
        """
        ang = (samples - low) * 2 * np.pi / (high - low)
        res = np.angle(np.mean(np.exp(1j * ang), axis=axis))
        mask = res < 0
        if mask.ndim > 0:
            res[mask] += 2 * np.pi
        elif mask:
            res += 2 * np.pi
        return res * (high - low) / 2.0 / np.pi + low

    @staticmethod
    def circstd(samples, high=2 * np.pi, low=0, axis=None):
        """Compute the circular standard deviation for samples assumed to be in the range [low to high].

        Taken from scipy.stats, with a small change on the 4th line.

        This uses a definition of circular standard deviation that in the limit of
        small angles returns a number close to the 'linear' standard deviation.

        Args:
            samples (array_like): Input array.
            low (float or int): Low boundary for circular standard deviation range.  Default is 0.
            high (float or int): High boundary for circular standard deviation range. Default is ``2*pi``.
            axis (int): Axis along which standard deviations are computed.  The default is
                to compute the standard deviation of the flattened array.

        Returns:
            float: Circular standard deviation.
        """
        ang = (samples - low) * 2 * np.pi / (high - low)
        res = np.mean(np.exp(1j * ang), axis=axis)
        R = abs(res)
        R[R >= 1] = 1 - np.finfo(np.float).eps
        return ((high - low) / 2.0 / np.pi) * np.sqrt(-2 * np.log(R))



import numpy as np

from mot.cl_data_type import CLDataType
from mot.data_adapters import SimpleDataAdapter
from mot.model_interfaces import OptimizeModelInterface

__author__ = 'Robbert Harms'
__date__ = "2015-04-02"
__license__ = "LGPL v3"
__maintainer__ = "Robbert Harms"
__email__ = "robbert.harms@maastrichtuniversity.nl"


class Rosenbrock(OptimizeModelInterface):

    def __init__(self, n=5):
        """When optimized the parameters should all be equal to 1."""
        super(Rosenbrock, self).__init__()
        self.n = n

    def double_precision(self):
        return True

    @property
    def name(self):
        return 'rosenbrock'

    def get_problems_var_data(self):
        return {}

    def get_problems_protocol_data(self):
        return {}

    def get_model_data(self):
        return {}

    def get_nmr_problems(self):
        return 1

    def get_model_eval_function(self, fname='evaluateModel'):
        return '''
            double ''' + fname + '''(const optimize_data* const data, const double* const x,
                                     const int observation_index){
                double sum = 0;
                for(int i = 0; i < ''' + str(self.n) + ''' - 1; i++){
                    sum += 100 * pown((x[i + 1] - pown(x[i], 2)), 2) + pown((x[i] - 1), 2);
                }
                return -sum;
            }
        '''

    def get_observation_return_function(self, fname='getObservation'):
        return '''
            double ''' + fname + '''(const optimize_data* const data, const int observation_index){
                return 0;
            }
        '''

    def get_objective_function(self, fname="calculateObjective"):
        eval_fname = fname + '_evaluateModel'
        obs_fname = fname + '_getObservation'
        func = self.get_model_eval_function(eval_fname)
        func += self.get_observation_return_function(obs_fname)
        return func + '''
            double ''' + fname + '''(const optimize_data* const data, double* const x){
                return ''' + obs_fname + '''(data, 1) -''' + eval_fname + '''(data, x, 1);
            }
        '''

    def get_objective_list_function(self, fname="calculateObjectiveList"):
        eval_fname = fname + '_evaluateModel'
        obs_fname = fname + '_getObservation'
        func = self.get_model_eval_function(eval_fname)
        func += self.get_observation_return_function(obs_fname)
        return func + '''
            void ''' + fname + '''(const optimize_data* const data, mot_float_type* const x, mot_float_type* result){
                result[1] = ''' + obs_fname + '''(data, 1) - ''' + eval_fname + '''(data, x, 1);
            }
        '''

    def get_initial_parameters(self, results_dict=None):
        params = np.ones((1, self.n)) * 3
        if results_dict:
            for i in range(self.n):
                if i in results_dict:
                    params[0, i] = results_dict[i]
        return SimpleDataAdapter(params, CLDataType.from_string('double'),
                                 CLDataType.from_string('double')).get_opencl_data()

    def get_lower_bounds(self):
        return ['-inf'] * self.n

    def get_upper_bounds(self):
        return ['inf'] * self.n

    def get_optimized_param_names(self):
        return range(self.n)

    def get_nmr_inst_per_problem(self):
        return 1

    def get_nmr_estimable_parameters(self):
        return self.n

    def get_parameter_codec(self):
        return None

    def get_final_parameter_transformations(self, fname='applyFinalParameterTransformations'):
        return None

    def finalize_optimization_results(self, results_dict):
        return results_dict


class MatlabLSQNonlinExample(OptimizeModelInterface):

    def __init__(self):
        """When optimized the parameters should be close to [0.2578, 0.2578] or something with a similar 2 norm.

        See the matlab manual page at http://nl.mathworks.com/help/optim/ug/lsqnonlin.html for more information.
        (viewed at 2015-04-02).

        """
        super(MatlabLSQNonlinExample, self).__init__()

    def double_precision(self):
        return True

    @property
    def name(self):
        return 'matlab_lsqnonlin_example'

    def get_problems_var_data(self):
        return {}

    def get_problems_protocol_data(self):
        return {}

    def get_model_data(self):
        return {}

    def get_nmr_problems(self):
        return 1

    def get_model_eval_function(self, fname='evaluateModel'):
        return '''
            double ''' + fname + '''(const optimize_data* const data, const double* const x,
                                     const int k){
                return -(2 + 2 * (k+1) - exp((k+1) * x[0]) - exp((k+1) * x[1]));
            }
        '''

    def get_observation_return_function(self, fname='getObservation'):
        return '''
            double ''' + fname + '''(const optimize_data* const data, const int observation_index){
                return 0;
            }
        '''

    def get_objective_function(self, fname="calculateObjective"):
        eval_fname = fname + '_evaluateModel'
        obs_fname = fname + '_getObservation'
        func = self.get_model_eval_function(eval_fname)
        func += self.get_observation_return_function(obs_fname)
        return func + '''
            double ''' + fname + '''(const optimize_data* const data, double* const x){
                double sum = 0;
                for(int i = 0; i < 10; i++){
                    sum += ''' + obs_fname + '''(data, i) - ''' + eval_fname + '''(data, x, i);
                }
                return sum;
            }
        '''

    def get_objective_list_function(self, fname="calculateObjectiveList"):
        eval_fname = fname + '_evaluateModel'
        obs_fname = fname + '_getObservation'
        func = self.get_model_eval_function(eval_fname)
        func += self.get_observation_return_function(obs_fname)
        return func + '''
            void ''' + fname + '''(const optimize_data* const data, mot_float_type* const x,
                                     mot_float_type* result){
                for(int i = 0; i < 10; i++){
                    result[i] = ''' + obs_fname + '''(data, i) - ''' + eval_fname + '''(data, x, i);
                }
            }
        '''

    def get_initial_parameters(self, results_dict=None):
        params = np.array([[0.3, 0.4]])
        if results_dict:
            for i in range(2):
                if i in results_dict:
                    params[0, i] = results_dict[i]
        return SimpleDataAdapter(params, CLDataType.from_string('double'),
                                 CLDataType.from_string('double')).get_opencl_data()

    def get_lower_bounds(self):
        return [0, 0]

    def get_upper_bounds(self):
        return ['inf', 'inf']

    def get_optimized_param_names(self):
        return [0, 1]

    def get_nmr_inst_per_problem(self):
        return 10

    def get_nmr_estimable_parameters(self):
        return 2

    def get_parameter_codec(self):
        return None

    def get_final_parameter_transformations(self, fname='applyFinalParameterTransformations'):
        return None

    def finalize_optimization_results(self, results_dict):
        return results_dict



from pkg_resources import resource_filename
from mot.cl_data_type import CLDataType
from mot.model_building.cl_functions.base import LibraryFunction
from mot.model_building.cl_functions.parameters import LibraryParameter

__author__ = 'Robbert Harms'
__date__ = "2016-10-03"
__maintainer__ = "Robbert Harms"
__email__ = "robbert.harms@maastrichtuniversity.nl"


class FirstLegendreTerm(LibraryFunction):

    def __init__(self):
        """A function for finding the first legendre term. (see the CL code for more details)
        """
        super(FirstLegendreTerm, self).__init__(
            'double',
            'getFirstLegendreTerm',
            (LibraryParameter(CLDataType.from_string('double'), 'x'),
             LibraryParameter(CLDataType.from_string('int'), 'n')),
            resource_filename('mot', 'data/opencl/firstLegendreTerm.h'),
            resource_filename('mot', 'data/opencl/firstLegendreTerm.cl'),
            {},
            ())


class Bessel(LibraryFunction):

    def __init__(self):
        """Function library for the bessel functions.

        See the CL code for more details.
        """
        super(Bessel, self).__init__(
            'mot_float_type',
            '',
            (),
            resource_filename('mot', 'data/opencl/bessel.h'),
            resource_filename('mot', 'data/opencl/bessel.cl'),
            {},
            ())


class CerfImWOfX(LibraryFunction):

    def __init__(self):
        """Calculate the cerf. (see the CL code for more details)
        """
        super(CerfImWOfX, self).__init__(
            'mot_float_type',
            'im_w_of_x',
            (LibraryParameter(CLDataType.from_string('mot_float_type'), 'x'),),
            resource_filename('mot', 'data/opencl/cerf/im_w_of_x.h'),
            resource_filename('mot', 'data/opencl/cerf/im_w_of_x.cl'),
            {},
            ())


class CerfDawson(LibraryFunction):

    def __init__(self):
        """Evaluate dawson integral. (see the CL code for more details)
        """
        super(CerfDawson, self).__init__(
            'mot_float_type',
            'dawson',
            (LibraryParameter(CLDataType.from_string('mot_float_type'), 'x'),),
            resource_filename('mot', 'data/opencl/cerf/dawson.h'),
            resource_filename('mot', 'data/opencl/cerf/dawson.cl'),
            {},
            (CerfImWOfX(),))


class CerfErfi(LibraryFunction):

    def __init__(self):
        """Calculate erfi. (see the CL code for more details)
        """
        super(CerfErfi, self).__init__(
            'mot_float_type',
            'erfi',
            (LibraryParameter(CLDataType.from_string('mot_float_type'), 'x'),),
            resource_filename('mot', 'data/opencl/cerf/erfi.h'),
            resource_filename('mot', 'data/opencl/cerf/erfi.cl'),
            {},
            (CerfImWOfX(),))


class EuclidianNormFunction(LibraryFunction):

    def __init__(self, memspace='private', memtype='mot_float_type'):
        """A CL functions for calculating the Euclidian distance between n values.

        Args:
            memspace (str): The memory space of the memtyped array (private, constant, global).
            memtype (str): the memory type to use, double, float, mot_float_type, ...
        """
        super(EuclidianNormFunction, self).__init__(
            memtype,
            'euclidian_norm_' + memspace + '_' + memtype,
            (LibraryParameter(CLDataType.from_string(memtype + '*'), 'x'),
             LibraryParameter(CLDataType.from_string('int'), 'n')),
            resource_filename('mot', 'data/opencl/euclidian_norm.ph'),
            resource_filename('mot', 'data/opencl/euclidian_norm.pcl'),
            {'MEMSPACE': memspace, 'MEMTYPE': memtype},
            ())



from collections import MutableMapping

import pyopencl as cl
import numpy as np
from ...utils import ParameterCLCodeGenerator, get_float_type_def
from ...cl_routines.base import CLRoutine
from ...load_balance_strategies import Worker


__author__ = 'Robbert Harms'
__date__ = "2014-02-05"
__license__ = "LGPL v3"
__maintainer__ = "Robbert Harms"
__email__ = "robbert.harms@maastrichtuniversity.nl"


class LogLikelihoodCalculator(CLRoutine):

    def calculate(self, model, parameters, evaluation_model=None):
        """Calculate and return the log likelihood of the given model under the given parameters.

        Args:
            model (AbstractModel): The model to calculate the full log likelihood for.
            parameters (dict or ndarray): The parameters to use in the evaluation of the model
                If a dict is given we assume it is with values for a set of parameters
                If an ndarray is given we assume that we have data for all parameters.
            evaluation_model (EvaluationModel): the evaluation model to use for the log likelihood. If not given
                we use the one defined in the model.

        Returns:
            Return per voxel the log likelihood.
        """
        parameters = self._initialize_parameters(parameters, model)
        log_likelihoods = self._initialize_result_array(model)

        workers = self._create_workers(
            lambda cl_environment: _LogLikelihoodCalculatorWorker(cl_environment, self.get_compile_flags_list(),
                                                                  model, parameters,
                                                                  log_likelihoods, evaluation_model))
        self.load_balancer.process(workers, model.get_nmr_problems())

        return log_likelihoods

    def _initialize_parameters(self, parameters, model):
        np_dtype = np.float32
        if model.double_precision:
            np_dtype = np.float64

        if isinstance(parameters, MutableMapping):
            return np.require(model.get_initial_parameters(parameters), np_dtype, requirements=['C', 'A', 'O'])

        return np.require(parameters, np_dtype, requirements=['C', 'A', 'O'])

    def _initialize_result_array(self, model):
        nmr_problems = model.get_nmr_problems()
        np_dtype = np.float32
        if model.double_precision:
            np_dtype = np.float64
        return np.zeros((nmr_problems,), dtype=np_dtype, order='C')


class _LogLikelihoodCalculatorWorker(Worker):

    def __init__(self, cl_environment, compile_flags, model, parameters, log_likelihoods, evaluation_model):
        super(_LogLikelihoodCalculatorWorker, self).__init__(cl_environment)

        self._model = model
        self._double_precision = model.double_precision
        self._log_likelihoods = log_likelihoods
        self._parameters = parameters
        self._evaluation_model = evaluation_model

        self._var_data_dict = model.get_problems_var_data()
        self._protocol_data_dict = model.get_problems_protocol_data()
        self._model_data_dict = model.get_model_data()

        self._all_buffers, self._likelihoods_buffer = self._create_buffers()
        self._kernel = self._build_kernel(compile_flags)

    def __del__(self):
        list(buffer.release() for buffer in self._all_buffers)

    def calculate(self, range_start, range_end):
        nmr_problems = range_end - range_start
        event = self._kernel.run_kernel(self._cl_run_context.queue, (int(nmr_problems), ), None, *self._all_buffers,
                                        global_offset=(int(range_start),))
        return [self._enqueue_readout(self._likelihoods_buffer, self._log_likelihoods, range_start, range_end, [event])]

    def _create_buffers(self):
        constant_buffers = self._generate_constant_buffers(self._protocol_data_dict, self._model_data_dict)

        likelihoods_buffer = cl.Buffer(self._cl_run_context.context,
                                       cl.mem_flags.WRITE_ONLY | cl.mem_flags.USE_HOST_PTR,
                                       hostbuf=self._log_likelihoods)

        params_buffer = cl.Buffer(self._cl_run_context.context,
                                  cl.mem_flags.READ_ONLY | cl.mem_flags.USE_HOST_PTR,
                                  hostbuf=self._parameters)

        var_data_buffers = []
        for data in self._var_data_dict.values():
            var_data_buffers.append(cl.Buffer(self._cl_run_context.context,
                                              cl.mem_flags.READ_ONLY | cl.mem_flags.USE_HOST_PTR,
                                              hostbuf=data.get_opencl_data()))

        all_buffers = [params_buffer, likelihoods_buffer]
        all_buffers.extend(var_data_buffers)
        all_buffers.extend(constant_buffers)

        return all_buffers, likelihoods_buffer

    def _get_kernel_source(self):
        cl_func = self._model.get_log_likelihood_function('getLogLikelihood', evaluation_model=self._evaluation_model)
        nmr_params = self._parameters.shape[1]

        param_code_gen = ParameterCLCodeGenerator(self._cl_environment.device, self._var_data_dict,
                                                  self._protocol_data_dict, self._model_data_dict)

        kernel_param_names = ['global mot_float_type* params', 'global mot_float_type* log_likelihoods']
        kernel_param_names.extend(param_code_gen.get_kernel_param_names())
        kernel_source = ''
        kernel_source += get_float_type_def(self._double_precision)
        kernel_source += param_code_gen.get_data_struct()
        kernel_source += cl_func
        kernel_source += '''
            __kernel void run_kernel(
                ''' + ",\n".join(kernel_param_names) + '''
                ){
                    int gid = get_global_id(0);
                    ''' + param_code_gen.get_data_struct_init_assignment('data') + '''

                    mot_float_type x[''' + str(nmr_params) + '''];
                    for(int i = 0; i < ''' + str(nmr_params) + '''; i++){
                        x[i] = params[gid * ''' + str(nmr_params) + ''' + i];
                    }

                    log_likelihoods[gid] = getLogLikelihood(&data, x);
            }
        '''
        return kernel_source



from mot.cl_routines.filters.base import AbstractFilter, AbstractFilterWorker
from mot.utils import get_float_type_def


__author__ = 'Robbert Harms'
__date__ = "2014-04-26"
__license__ = "LGPL v3"
__maintainer__ = "Robbert Harms"
__email__ = "robbert.harms@maastrichtuniversity.nl"


class MeanFilter(AbstractFilter):

    def _get_worker_generator(self, *args):
        return lambda cl_environment: _MeanFilterWorker(cl_environment, self.get_compile_flags_list(), *args)


class _MeanFilterWorker(AbstractFilterWorker):

    def _get_kernel_source(self):
        kernel_source = ''
        kernel_source += get_float_type_def(self._double_precision)
        kernel_source += self._get_ks_sub2ind_func(self._volume_shape)
        kernel_source += '''
            __kernel void filter(
                global mot_float_type* volume,
                ''' + ('global char* mask,' if self._use_mask else '') + '''
                global mot_float_type* results
                ){
                    ''' + self._get_ks_dimension_inits(len(self._volume_shape)) + '''
                    int ind = ''' + self._get_ks_sub2ind_func_call(len(self._volume_shape)) + ''';

                    ''' + ('if(mask[ind] > 0){' if self._use_mask else 'if(true){') + '''
                        ''' + self._get_ks_dimension_sizes(self._volume_shape) + '''

                        mot_float_type sum = 0.0;
                        int count = 0;

                        ''' + self._get_ks_loop(self._volume_shape) + '''

                        results[ind] = sum/count;
                    }
            }
        '''
        return kernel_source

    def _get_ks_loop(self, volume_shape):
        s = ''
        for i in range(len(volume_shape)):
            s += 'for(dim' + str(i) + ' = dim' + str(i) + '_start; dim' + str(i) + \
                 ' < dim' + str(i) + '_end; dim' + str(i) + '++){' + "\n"

        if self._use_mask:
            s += 'if(mask[' + self._get_ks_sub2ind_func_call(len(volume_shape)) + '] > 0){' + "\n"

        s += 'sum += volume[' + self._get_ks_sub2ind_func_call(len(volume_shape)) + '];' + "\n"
        s += 'count++;' + "\n"

        if self._use_mask:
            s += '}' + "\n"

        for i in range(len(volume_shape)):
            s += '}' + "\n"
        return s



import pyopencl as cl
from ...utils import results_to_dict, ParameterCLCodeGenerator, get_float_type_def
from ...cl_routines.base import CLRoutine
from ...load_balance_strategies import Worker
import numpy as np


__author__ = 'Robbert Harms'
__date__ = "2014-02-05"
__license__ = "LGPL v3"
__maintainer__ = "Robbert Harms"
__email__ = "robbert.harms@maastrichtuniversity.nl"


class CalculateDependentParameters(CLRoutine):

    def __init__(self, cl_environments=None, load_balancer=None, compile_flags=None, double_precision=False, **kwargs):
        """CL code for calculating the dependent parameters.

        Some of the models may contain parameter dependencies. We would like to return the maps for these parameters
        as well as all the other maps. Since the dependencies are specified in CL, we have to recourse to CL to
        calculate these maps.

        Args:
            double_precision (boolean): if we will use the double (True) or single floating (False) type
                for the calculations
        """
        super(CalculateDependentParameters, self).__init__(cl_environments=cl_environments,
                                                           load_balancer=load_balancer, **kwargs)
        self._double_precision = double_precision

    def calculate(self, fixed_param_values, estimated_parameters_list, parameters_listing, dependent_parameter_names):
        """Calculate the dependent parameters

        This uses the calculated parameters in the results dictionary to run the parameters_listing in CL to obtain
        the maps for the dependent parameters.

        Args:
            fixed_param_values (dict): The dictionary with fixed parameter values to be used for parameters that are
                constant. The values are supposed to be DataAdapters. This is interpreted as 'var_data_dict'.
                Data that is variable per problem instance.
            estimated_parameters_list (list of ndarray): The list with the one-dimensional
                ndarray of estimated parameters
            parameters_listing (str): The parameters listing in CL
            dependent_parameter_names (list of list of str): Per parameter we would like to obtain the CL name and the
                result map name. For example: (('Wball_w', 'Wball.w'),)
        Returns:
            dict: A dictionary with the calculated maps for the dependent parameters.
        """
        np_dtype = np.float32
        if self._double_precision:
            np_dtype = np.float64

        results_list = np.zeros(
            (estimated_parameters_list[0].shape[0], len(dependent_parameter_names)),
            dtype=np_dtype, order='C')

        estimated_parameters = np.require(np.dstack(estimated_parameters_list),
                                          np_dtype, requirements=['C', 'A', 'O'])

        workers = self._create_workers(
            lambda cl_environment: _CDPWorker(cl_environment, self.get_compile_flags_list(),
                                              fixed_param_values, len(estimated_parameters_list),
                                              estimated_parameters, parameters_listing,
                                              dependent_parameter_names, results_list, self._double_precision))
        self.load_balancer.process(workers, estimated_parameters_list[0].shape[0])

        return results_to_dict(results_list, [n[1] for n in dependent_parameter_names])


class _CDPWorker(Worker):

    def __init__(self, cl_environment, compile_flags, var_data_dict, nmr_estimated_params, estimated_parameters,
                 parameters_listing, dependent_parameter_names, results_list, double_precision):
        super(_CDPWorker, self).__init__(cl_environment)

        self._var_data_dict = var_data_dict
        self._nmr_estimated_params = nmr_estimated_params
        self._parameters_listing = parameters_listing
        self._dependent_parameter_names = dependent_parameter_names
        self._results_list = results_list
        self._double_precision = double_precision

        self._estimated_parameters = estimated_parameters
        self._all_buffers, self._results_list_buffer = self._create_buffers()
        self._kernel = self._build_kernel(compile_flags)

    def __del__(self):
        list(buffer.release() for buffer in self._all_buffers)

    def calculate(self, range_start, range_end):
        nmr_problems = int(range_end - range_start)

        event = self._kernel.transform(self._cl_run_context.queue, (int(nmr_problems), ), None, *self._all_buffers,
                                       global_offset=(int(range_start),))
        return [self._enqueue_readout(self._results_list_buffer, self._results_list, range_start, range_end, [event])]

    def _create_buffers(self):
        estimated_parameters_buf = cl.Buffer(self._cl_run_context.context,
                                             cl.mem_flags.READ_ONLY | cl.mem_flags.USE_HOST_PTR,
                                             hostbuf=self._estimated_parameters)

        results_buffer = cl.Buffer(self._cl_run_context.context,
                                   cl.mem_flags.WRITE_ONLY | cl.mem_flags.USE_HOST_PTR,
                                   hostbuf=self._results_list)

        data_buffers = [estimated_parameters_buf, results_buffer]

        for data in self._var_data_dict.values():
            data_buffers.append(cl.Buffer(self._cl_run_context.context,
                                          cl.mem_flags.READ_ONLY | cl.mem_flags.COPY_HOST_PTR,
                                          hostbuf=data.get_opencl_data()))

        return data_buffers, results_buffer

    def _get_kernel_source(self):
        dependent_parameter_names = [n[0] for n in self._dependent_parameter_names]

        parameter_write_out = ''
        for i, p in enumerate(dependent_parameter_names):
            parameter_write_out += 'results[gid * ' + str(len(dependent_parameter_names)) + \
                                   ' + ' + str(i) + '] = ' + p + ";\n"

        param_code_gen = ParameterCLCodeGenerator(self._cl_environment.device, self._var_data_dict, {}, {})
        kernel_param_names = ['global mot_float_type* params', 'global mot_float_type* results']
        kernel_param_names.extend(param_code_gen.get_kernel_param_names())

        kernel_source = ''
        kernel_source += get_float_type_def(self._double_precision)
        kernel_source += param_code_gen.get_data_struct()
        kernel_source += '''
            __kernel void transform(
                ''' + ",\n".join(kernel_param_names) + '''
                ){
                    int gid = get_global_id(0);

                    ''' + param_code_gen.get_data_struct_init_assignment('data_var') + '''
                    optimize_data* data = &data_var;

                    mot_float_type x[''' + str(self._nmr_estimated_params) + '''];
                    int i = 0;
                    for(i = 0; i < ''' + str(self._nmr_estimated_params) + '''; i++){
                        x[i] = params[gid * ''' + str(self._nmr_estimated_params) + ''' + i];
                    }
                    ''' + self._parameters_listing + '''
                    ''' + parameter_write_out + '''
            }
        '''
        return kernel_source



import logging
import pyopencl as cl
import numpy as np
from ...utils import get_float_type_def
from ...cl_routines.base import CLRoutine
from ...load_balance_strategies import Worker


__author__ = 'Robbert Harms'
__date__ = "2014-05-18"
__license__ = "LGPL v3"
__maintainer__ = "Robbert Harms"
__email__ = "robbert.harms@maastrichtuniversity.nl"


class CodecRunner(CLRoutine):

    def __init__(self, cl_environments=None, load_balancer=None, double_precision=False):
        """This class can run the codecs used to transform the parameters to and from optimization space.

        Args:
            double_precision (boolean): if we will use the double (True) or single floating (False) type for the calculations
        """
        super(CodecRunner, self).__init__(cl_environments, load_balancer)
        self._logger = logging.getLogger(__name__)
        self._double_precision = double_precision

    def decode(self, codec, data):
        """Decode the parameters.

        This transforms the data from optimization space to model space.

        Args:
            codec (AbstractCodec): The codec to use in the transformation.
            data (ndarray): The parameters to transform to model space

        Returns:
            ndarray: The array with the transformed parameters.
        """
        if len(data.shape) > 1:
            from_width = data.shape[1]
        else:
            from_width = 1

        if from_width != codec.get_nmr_parameters():
            raise ValueError("The width of the given data does not match the codec expected width.")

        return self._transform_parameters(codec.get_cl_decode_function('decodeParameters'), 'decodeParameters', data,
                                          codec.get_nmr_parameters())

    def encode(self, codec, data):
        """Encode the parameters.

        This transforms the data from model space to optimization space.

        Args:
            codec (AbstractCodec): The codec to use in the transformation.
            data (ndarray): The parameters to transform to optimization space

        Returns:
            ndarray: The array with the transformed parameters.
        """
        if len(data.shape) > 1:
            from_width = data.shape[1]
        else:
            from_width = 1

        if from_width != codec.get_nmr_parameters():
            raise ValueError("The width of the given data does not match the codec expected width.")

        return self._transform_parameters(codec.get_cl_encode_function('encodeParameters'), 'encodeParameters', data,
                                          codec.get_nmr_parameters())

    def _transform_parameters(self, cl_func, cl_func_name, data, nmr_params):
        np_dtype = np.float32
        if self._double_precision:
            np_dtype = np.float64
        data = np.require(data, np_dtype, requirements=['C', 'A', 'O', 'W'])
        rows = data.shape[0]
        workers = self._create_workers(lambda cl_environment: _CodecWorker(cl_environment, self.get_compile_flags_list(),
                                                                           cl_func, cl_func_name, data,
                                                                           nmr_params, self._double_precision))
        self.load_balancer.process(workers, rows)
        return data


class _CodecWorker(Worker):

    def __init__(self, cl_environment, compile_flags, cl_func, cl_func_name, data, nmr_params, double_precision):
        super(_CodecWorker, self).__init__(cl_environment)
        self._cl_func = cl_func
        self._cl_func_name = cl_func_name
        self._data = data
        self._nmr_params = nmr_params
        self._double_precision = double_precision

        self._param_buf = cl.Buffer(self._cl_run_context.context,
                                    cl.mem_flags.READ_WRITE | cl.mem_flags.USE_HOST_PTR,
                                    hostbuf=self._data)

        self._kernel = self._build_kernel(compile_flags)

    def calculate(self, range_start, range_end):
        nmr_problems = range_end - range_start

        event = self._kernel.transformParameterSpace(self._cl_run_context.queue, (int(nmr_problems), ), None,
                                                     self._param_buf, global_offset=(int(range_start),))

        return [self._enqueue_readout(self._param_buf, self._data, range_start, range_end, [event])]

    def _get_kernel_source(self):
        kernel_source = ''
        kernel_source += get_float_type_def(self._double_precision)
        kernel_source += self._cl_func
        kernel_source += '''
            __kernel void transformParameterSpace(global mot_float_type* x_global){
                int gid = get_global_id(0);

                mot_float_type x[''' + str(self._nmr_params) + '''];

                for(int i = 0; i < ''' + str(self._nmr_params) + '''; i++){
                    x[i] = x_global[gid * ''' + str(self._nmr_params) + ''' + i];
                }

                ''' + self._cl_func_name + '''(x);

                for(int i = 0; i < ''' + str(self._nmr_params) + '''; i++){
                    x_global[gid * ''' + str(self._nmr_params) + ''' + i] = x[i];
                }
            }
        '''
        return kernel_source



__author__ = 'Robbert Harms'
__date__ = "2014-06-19"
__license__ = "LGPL v3"
__maintainer__ = "Robbert Harms"
__email__ = "robbert.harms@maastrichtuniversity.nl"


class AbstractParameterPrior(object):
    """The priors are used during model sampling.

    They indicate the a priori information one has about a parameter.
    """

    def get_cl_assignment(self, parameter, parameter_name):
        """Get the assignment code.

        In CL assignments look like: a = b;
        This function should return the "b;" part. That is, all that follows after the assignment operator.

        Args:
            parameter (FreeParameter): the free parameter for which to create a prior
            parameter_name (str): the name of the parameter for which we create the prior

        Returns:
            str: The assignment code for the prior
        """
        return ''


class UniformPrior(AbstractParameterPrior):
    """The uniform prior is always 1."""
    def get_cl_assignment(self, parameter, parameter_name):
        return '1;'


class UniformWithinBoundsPrior(AbstractParameterPrior):
    """This prior is 1 within the upper and lower bound of the parameter, 0 outside."""
    def get_cl_assignment(self, parameter, parameter_name):
        return '((' + parameter_name + ' < ' + str(float(parameter.lower_bound)) + \
               ' || ' + parameter_name + ' > ' + str(float(parameter.upper_bound)) + ') ? 0.0 : 1.0);'


class AbsSinPrior(AbstractParameterPrior):
    """The fabs(sin(x)) prior."""
    def get_cl_assignment(self, parameter, parameter_name):
        return 'fabs(sin(' + parameter_name + '));'


class AbsSinHalfPrior(AbstractParameterPrior):
    """The fabs(sin(x)/2.0) prior. Taken from FSL"""
    def get_cl_assignment(self, parameter, parameter_name):
        return 'fabs(sin(' + parameter_name + ')/2.0);'
